--
--	Wholly
--	Written by scott@mithrandir.com
--
--	This was inspired by my need to replace EveryQuest with something that was more accurate.  The pins
--	was totally inspired by QuestHubber which I found when looking for a replacement for EveryQuest.  I
--	initially made a port of QuestHubber to QuestHubberGrail to make use of the Grail database, but now
--	I have integrated that work into this addon.  Many thanks for all the work put into QuestHubber.  I
--	was inspired to add a quest breadcrumb area from seeing one in Quest Completist.
--
--	Version History
--		001	Initial version.
--		002	Support for displaysDungeonQuests now works properly.
--			Added the ability for the panel tooltip to display questgivers.
--			Added the ability to click a quest in the panel to create a TomTom waypoint.
--			Map pins are only displayed for the proper dungeon level of the map.
--		003	Added a panel button to switch to the current zone.
--			Changed the Close button into a Sort button that switches between three different modes:
--				1. Alphabetical  2. Quest level, then alphabetical  3. Quest status, then alphabetical
--			Made map pins only have one pin per NPC, indicating the "best" color possible.
--			The entire zone dropdown button on the quest log panel now can be clicked to change zones.
--			Corrected a problem where someone running without LibStub would get a LUA error.
--			Corrected a localization issue.
--		004	Added the ability to display a basic quest prerequisite chain in the quest panel tooltip, requiring Grail 014 or later.
--			Added the ability to right-click a "prerequisite" quest in the panel to put a TomTom arrow to the first prerequisite quest.
--			Added Dungeons and Other menu items for map areas in the quest log panel.
--			The last-used sort preference is stored on a per-character basis.
--		005	Corrected the fix introduced in version 003 putting the LibDataBroker icon back in place.
--			Corrected a problem where the quest log tooltip would have an error if the questgiver was in a dungeon appearing in the zone.
--			Added the ability for the quest log tooltip to show that the questgiver should be killed (like the map pin tooltip).
--			The problem where map pins would not live update unless the quest log panel was opened has been fixed as long
--				as the Wholly check button appears on the map.
--			Added the ability to show quest breadcrumb information when the Quest Frame opens, showing a tooltip of breadcrumb
--				quests when the mouse enters the "breadcrumb area", and putting TomTom waypoints when clicking in it.
--		006	Added the new quest group "World Events" which has holiday quests in it, requiring Grail 015.
--			Added a tooltip to the Wholly check button on the map that indicates how many quests of each type are in the map.
--			Added a tooltip to the LibDataBroker icon that shows the quest log panel "map" selection and the quest count/type.
--			Added a tooltip to the quest log panel Zone button that shows the quest count/type.
--			Corrected the problem where the quest log panel and map pins were not live updating when quest givers inside dungeons checked.
--			Corrected the problem where an NPC that drops items that starts more than one quest does not display the information properly
--				in its tooltip.
--			Made it so the open World Map can be updated when crossing into a new zone.
--		007	Added the ability to show whether quests in the quest log are completed or failed.
--			Made it so right-clicking an "in log" quest will put in TomTom waypoints for the turn in locations, which requires Grail 016
--				for proper functioning since Grail 015 had a bug.
--			Made the strings for the preferences color quest information like it appears in the UI.
--			Made it so alt-clicking a log in the Wholly quest log selects the NPC that gives the quest or for the case of "in log" quests
--				the one to which the quest is turned in.
--		008	Split out Dungeons into dungeons in different continents, requiring Grail version 017.
--			Corrected a misspelling of the global game tooltip name.
--		009	Added localization for ptBR in anticipation of the Brazilian release.
--			Changed over to using Blizzard-defined strings for as many things as possible.
--			Corrected a problem that was causing the tooltip for creatures that needed to be killed to start a quest not to appear properly.
--			Added a few basic localizations.
--			Made the breadcrumb frame hide by default to attempt to eliminate an edge case.
--			Fixed a problem where button clicking behavior would never be set if the button was first entered while in combat.
--			Made prerequisite information appear as question marks instead of causing a LUA error in case the Grail data is lacking.
--		010	Made it so the color of the breadcrumb quest names match their current status.
--			The click areas to the right and bottom of the quest log window no longer extend past the window.
--			Added menu options for Class quests, Profession quests, Reputation quests, and Daily quests.  The Class and Profession quests will show all the quests in the system except for the class and professions that match the player.  For those, the quests are displayed using the normal filtering rules.  The Reputation quests follow the normal filtering rules except those that fail to be acceptable solely because of reputation will be displayed instead of following the display unobtainable filter.
--			Changed over to using Grail's StatusCode() vice Status(), and making use of more modern bit values, thereby requiring version 20.
--			Removed a few event types that are handled because Grail now does that.  Instead switched to using Grail's new Status notification.
--			The tooltips for quests in the panel show profession and reputation requirements as appropriate.
--			Corrected a problem where the quest panel may not update properly while in combat.
--		011	Made it so the breadcrumb warning will disappear properly when the user dismisses the quest frame.
--			Made it so Grail's support for Monks does not cause headaches when Monks are not available in the game.
--			Made it so Classes that do not have any class quests will not show up in the list.
--			Put in a feature to limit quests shown to those that count towards Loremaster, thereby requiring Grail version 21.
--			When the quest details appear the quest ID is shown in the top right, and it has a tooltip with useful quest information.
--			Changed the behavior of right-clicking a quest in the quest panel to put arrows to the turn in locations for all but prerequisite quests.
--			The tooltip information describing the quest shows failure reasons by changing to red categories that fail, and to orange categories that fail in prerequisite quests.
--			The quest tooltip information now indicates the first quest(s) in the prerequisite quest list as appropriate.
--			The preference to control displaying prerequisite quests in the tooltip has been removed.
--		012	Added the ability for the tooltip to display faction reputation changes that happen when a quest is turned in.
--			Grouped the Dungeons menu items under a single Dungeons menu item.
--			Added menu items for Reputation Changes quests, grouped by reputation.
--			Added menu items for Achievement quests, grouped by continent, requiring Grail 22.
--			Updated the TOC to support Interface 40300.
--			Fixed the map pin icons whose height Blizzard changed with the 4.3 release.
--		013	Fixes a problem where map pins would not appear and the quest ID would not appear in the Blizzard Quest Frame because the events were not set up properly because sometimes Blizzard sends events in a different order than expected.
--			Makes all the Wholly quest panel update calls ensure they are performed out of combat.
--			Updates Portuguese translation thanks to weslleih and htgome.
--			Fixes a problem where quests in the Blizzard log sometimes would not appear purple in the Wholly Quest Log.
--			Fixes a problem where holidays are not detected properly because of timing issues.
--		014	Fixes the problem where the NPC tooltip did not show the exclamation point properly (instead showing a different icon) when the NPC can start a quest.
--			Adds a search ability that allows searching for quests based on their titles.
--			Adds the ability to display player coordinates into a LibDataBroker feed.
--			Updates some localizations.
--			Fixes the problem where the panel would no longer update after a UI reload, requiring Grail 26.
--			Adds some more achievements to the menu that are world event related.
--			Makes it so quests in the Blizzard quest log will be colored purple in preference to other colors (like brown in case the player would no longer qualify getting the quest).
--			Makes it so the indicator for a completed repeatable quest will appear even if the quest is not colored blue.
--		015	Adds the filtered and total quest counts to the tooltip that tells the counts of the types of quests.  For the world map button tooltip the filtered quest count displays in red if the quest markers on the map are hidden.
--			Corrects a problem where lacking LibDataBroker would cause a LUA error associated with the player coordinates.
--			Fixes a cosmetic issue with the icon in the top left of the Wholly quest log panel to show the surrounding edge properly.
--			Changes the world map check box into a button that performs the same function.
--			Changes the classification of "weekly", "monthly" and "yearly" quests so they no longer appear as resettable quests, but as normal ones.
--			Adds a tooltip for the coordinates that shows the map area ID and name.
--
--	Known Issues
--			When adding TomTom arrows, it may be useful to be able to add all the waypoints as a group so they can be removed together.  For example, adding all the commoners together and removing when one of them is found.  This is not ceonveniently done because it means Wholly has to keep them grouped together, and then be notified when one is removed because it has been reached.  To set up the callback for this one loses all the basic TomTom default settings for other aspects that are handled by callbacks.
--			When an item that drops a quest has multiple NPCs that have it, Wholly is showing the same name over and over instead of each different one.  This is because Grail:LocationNPCInMap cannot return more than one NPC name for the list of locations.  Therefore, for the data to appear accurately in the tooltip, Grail needs to be able to provide this information correctly.
--			When a quest starts with an NPC that drops something the name of the NPC includes that something which makes targeting it impossible.
--
--	UTF-8 file
--

local Wholly_File_Version = 015
local requiredGrailVersion = 26

--	Set up the bindings to use the localized name Blizzard supplies.  Note that the Bindings.xml file cannot
--	just contain the TOGGLEQUESTLOG because then the entry for Wholly does not show up.  So, we use a version
--	named WHOLLY_TOGGLEQUESTLOG which maps to the same Global string, which works exactly as we want.
BINDING_NAME_WHOLLY_TOGGLEQUESTLOG = BINDING_NAME_TOGGLEQUESTLOG
BINDING_HEADER_WHOLLY = "Wholly"

if nil == Wholly or Wholly.versionNumber < Wholly_File_Version then

	WhollyDatabase = {}

	Wholly = {

		cachedPanelQuests = {},		-- quests and their status for map area currentPanelMapID
		cachedPinQuests = {},		-- quests and their status for map area currentPinMapID
		color = {
			['B'] = "FF996600",	-- brown	[unobtainable]
			['C'] = "FF00FF00",	-- green	[completed]
			['D'] = "FF0099CC",	-- daily	[repeatable]
			['G'] = "FFFFFF00",	-- yellow	[can accept]
			['I'] = "FFFF00FF",	-- purple	[in log]
			['L'] = "FFFFFFFF",	-- white	[too high level]
			['P'] = "FFFF0000",	-- red		[does not meet prerequisites]
			['R'] = "FF0099CC",	-- daily	[true repeatable - used for question mark in pins]
			['U'] = "FF00FFFF",	-- bogus default[unknown]
			['W'] = "FF666666",	-- grey		[low-level can accept]
			},
		configuration = {
			[1] = {[0]="Prepend quest level", [1]='prependsQuestLevel' },
			[2] = {[0]="Append required level", [1]='appendRequiredLevel' },
			[3] = {[0]="Show |cFF0099CCrepeatable quests|r", [1]='showsDailyQuests' },
			[4] = {[0]="Show |cFFFF00FFquests in log|r", [1]='showsQuestsInLog' },
			[5] = {[0]="Show |cFF00FF00completed quests|r", [1]='showsCompletedQuests' },
			[6] = {[0]="Show |cFF996600unobtainable quests|r", [1]='showsUnobtainableQuests' },
			[7] = {[0]="Show |cFF666666low-level quests|r", [1]='showsLowLevelQuests' },
			[8] = {[0]="Show whether repeatable quests previously completed", [1]='showsAnyPreviousRepeatableCompletions' },
			[9] = {[0]="Show high-level quests", [1]='showsHighLevelQuests'},
			[10]= {[0]="Show |cFFFF0000quests that need prerequisites|r", [1]='showsQuestsThatFailPrerequsites'},
			[11]= {[0]="Quest log panel updates when zones change", [1]='updatesPanelWhenZoneChanges'},
			[12]= {[0]="Display map pins for quest givers", [1]='displaysMapPins'},
			[13]= {[0]="Display button on world map", [1]='displaysMapFrame'},
			[14]= {[0]="Display dungeon quests in outer map", [1]='displaysDungeonQuests'},
			[15]= {[0]="Display breadcrumb quest information on Quest Frame", [1]='displaysBreadcrumbs'},
			[16]= {[0]="Display holiday quests always", [1]='displaysHolidaysAlways'},
			[17]= {[0]="Open world map updates when zones change", [1]='updatesWorldMapOnZoneChange'},
			[18]= {[0]="Show status of quests in log", [1]='showsInLogQuestStatus'},
			[19]= {[0]="Show only Loremaster quests", [1]='showsLoremasterOnly'},
			[20]= {[0]="Enable player coordinates", [1]='enablesPlayerCoordinates'},
			},
		coordinates = nil,
		currentPanelMapID = nil,	-- the map area the panel is showing
		currentPinMapID = nil,		-- the map area the world map is showing
		currentPinMapLevel = 0,
		debug = true,
		dropdown = nil,
		dungeonTest = {},
		eventDispatch = {
			['PLAYER_REGEN_ENABLED'] = function(self, frame)
				self:ScrollFrame_Update()
				frame:UnregisterEvent("PLAYER_REGEN_ENABLED")
			end,
			--	So in Blizzard's infinite wisdom it turns out that normal quests that just appear with the
			--	quest giver post a QUEST_DETAIL event, unless they are quests like the Candy Bucket quests
			--	which post a QUEST_COMPLETE event (even though they really are not complete until they are
			--	accepted).  And if there are more than one quest then QUEST_GREETING is posted, which also
			--	is posted if one were to decline one of the selected ones to return to the multiple choice
			--	frame again.  Therefore, it seems three events are required to ensure the breadcrumb panel
			--	is properly removed.
			['QUEST_COMPLETE'] = function(self, frame)
				local questId = self:BreadcrumbQuestId()
				com_mithrandir_whollyQuestInfoFrameText:SetText(questId)
				com_mithrandir_whollyBreadcrumbFrame:Hide()
			end,
			['QUEST_DETAIL'] = function(self, frame)
				local questId = self:BreadcrumbQuestId()
				com_mithrandir_whollyQuestInfoFrameText:SetText(questId)
				self:ShowBreadcrumbInfo()
			end,
			['QUEST_GREETING'] = function(self, frame)
				com_mithrandir_whollyQuestInfoFrameText:SetText("")
				com_mithrandir_whollyBreadcrumbFrame:Hide()
			end,
			['QUEST_PROGRESS'] = function(self, frame)
				local questId = self:BreadcrumbQuestId()
				com_mithrandir_whollyQuestInfoFrameText:SetText(questId)
				com_mithrandir_whollyBreadcrumbFrame:Hide()
			end,
			['ADDON_LOADED'] = function(self, frame, arg1)
				if "Wholly" == arg1 then
					if nil == WhollyDatabase.defaultsLoaded then
						self:LoadDefaults()
					end
					if nil == WhollyDatabase.currentSortingMode then
						WhollyDatabase.currentSortingMode = 1
					end

					-- Now to be nicer to those that have used the addon before the current
					-- incarnation, newly added defaults will have their normal setting set
					-- as appropriate.
					if nil == WhollyDatabase.version then		-- first loaded prior to version 006, so default options added in 006
						WhollyDatabase.displaysHolidaysAlways = true		-- version 006
						WhollyDatabase.updatesWorldMapOnZoneChange = true	-- version 006
					elseif WhollyDatabase.version < 7 then
						WhollyDatabase.showsInLogQuestStatus = true			-- version 007
					end
					WhollyDatabase.version = Wholly.versionNumber

					self:DisplayMapFrame(WhollyDatabase.displaysMapFrame)
					Grail:RegisterObserver("Status", self.CallbackHandler)
					Grail:RegisterObserverQuestAbandon(self.CallbackHandler)

					-- Find out which "map area" is for the player's class
					for key, value in pairs(Grail.classMapping) do
						if Grail.playerClass == value then
							self.playerClassMap = Grail.classToMapAreaMapping['C'..key]
						end
					end

					self:UpdateCoordinateSystem()	-- installs OnUpdate script appropriately

					frame:RegisterEvent("QUEST_COMPLETE")			-- to clear the breadcrumb frame
					frame:RegisterEvent("QUEST_GREETING")			-- to clear the breadcrumb frame
					frame:RegisterEvent("QUEST_PROGRESS")
					frame:RegisterEvent("WORLD_MAP_UPDATE")			-- this is for pins
					frame:RegisterEvent("ZONE_CHANGED_NEW_AREA")	-- this is for the panel
					self:UpdateBreadcrumb()							-- sets up registration of events for breadcrumbs based on user preferences
				end
			end,
			['WORLD_MAP_UPDATE'] = function(self, frame)
				self:UpdateQuestCaches(false, true)	-- note that we do not force the update since this event happens way too many times
			end,
			['ZONE_CHANGED_NEW_AREA'] = function(self, frame)
				if WhollyDatabase.updatesWorldMapOnZoneChange and WorldMapFrame:IsVisible() then
					local zoneNameToUse = GetRealZoneText()
					if nil ~= zoneNameToUse then
						local newMapId = Grail.zoneNameMapping[zoneNameToUse]
						if nil ~= newMapId then
							SetMapByID(newMapId)
						end
					end
				end
				self:UpdateQuestCaches(false, false, WhollyDatabase.updatesPanelWhenZoneChanges)

				-- Now update open tooltips showing our quest count data
				if GameTooltip:IsVisible() and GameTooltip:GetOwner() == com_mithrandir_whollyFrameSwitchZoneButton then
					GameTooltip:ClearLines()
					GameTooltip:AddLine(Wholly.panelCountLine)
				elseif self.tooltip:IsVisible() and self.tooltip:GetOwner() == self.mapFrame then
					self.tooltip:ClearLines()
					self.tooltip:AddLine(Wholly.mapCountLine)
				elseif GameTooltip:IsVisible() and GameTooltip:GetOwner() == self.ldbTooltipOwner then -- LibDataBroker tooltip
					GameTooltip:ClearLines()
					GameTooltip:AddLine("Wholly - " .. UIDropDownMenu_GetText(Wholly.dropdown) )
					GameTooltip:AddLine(Wholly.panelCountLine)
				elseif GameTooltip:IsVisible() and GameTooltip:GetOwner() == self.ldbCoordinatesTooltipOwner then -- LibDataBroker coordinates tooltip
					GameTooltip:ClearLines()
					local dungeonLevel = GetCurrentMapDungeonLevel()
					local dungeonIndicator = (dungeonLevel > 0) and "["..dungeonLevel.."]" or ""
					local mapAreaId = GetCurrentMapAreaID()
					local mapAreaName = Grail.mapAreaMapping[mapAreaId] or "UNKNOWN"
					GameTooltip:AddLine(string.format("%d%s %s", mapAreaId, dungeonIndicator, mapAreaName))
				end
			end,
			},
		filteredPanelQuests = {},	-- filtered table from cachedPanelQuests using current panel filters
		filteredPinQuests = {},		-- filtered table from cachedPinQuests using current pin filters
		initialUpdateProcessed = false,
		lastPrerequisiteQuest = nil,
		lastUpdate = 0,
		ldbCoordinatesTooltipOwner = nil,
		ldbTooltipOwner = nil,
		mapFrame = nil,			-- the world map frame that contains the checkbox to toggle pins
		maximumSearchHistory = 10,
		npcs = {},
		pairedConfigurationButton = nil,-- configuration panel button that does the same thing as the world map button
		pairedCoordinatesButton = nil,	-- configuration panel button that does the same thing as the LDB coordinate button
		playerAliveReceived = false,
		playerClassMap = nil,
		previousX = 0,
		previousY = 0,
		receivedCalendarUpdateEventList = false,
		regFrames = {},
		regNPCPins = {},		-- in unifying the pins we keep a table of pins associated with NPC id
		regPins = {},
		reputationLevelMapping = { [0] = 1, [35999] = 2, [38999] = 3, [41999] = 4, [44999] = 5, [50999] = 6, [62999] = 7, [83999] = 8, [84998] = 8, },
		s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['PREQ_EVER'] = "Prerequisite Quests (Ever Completed):",
			['PREQ_LOG'] = "Prerequisite Quests (Must Be in Log):",
			['PREQ_LOG_TURN'] = "Prerequisite Quests (Must Be in Log or Turned in):",
			['PREQ_COMPLETE'] = "Prerequisite Quests (Must Be Completed in Log):",
			['PREQ_COMPLETE_TURN'] = "Prerequisite Quests (Must Be Completed in Log or Turned in):",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			},
		tooltip = nil,
		updateDelay = 0.5,
		updateThreshold = 0.1,
		versionNumber = Wholly_File_Version,

		AddDirectionalArrows = function(self, questTable, npcType)
			if not TomTom or not TomTom.AddMFWaypoint then return end
			if nil == questTable or nil == npcType then return end
			local locations, mustKill
			for _, questId in pairs(questTable) do
				locations, mustKill = Grail:MultipleUniqueMapLocationQuest(questId, npcType, nil, nil, true)
				if nil ~= locations then
					for i = 1, #locations do
						local xcoord, ycoord, npcname, npcId, mapId, dungeonLevel, realAreaId = locations[i][1], locations[i][2], locations[i][3], locations[i][4], locations[i][5], locations[i][6], locations[i][7]
						if xcoord ~= nil then
							npcname = self:PrettyNPCString(npcname, mustKill, realAreaId)
							TomTom:AddMFWaypoint(mapId, dungeonLevel, xcoord/100, ycoord/100, { title = npcname.." - "..self:QuestName(questId) })
						end
					end
				end
			end
		end,

		BreadcrumbClick = function(self, frame)
			local questId = self:BreadcrumbQuestId()
			self:AddDirectionalArrows(Grail:AvailableBreadcrumbs(questId), 'A')
		end,

		BreadcrumbEnter = function(self, frame)
			GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
			GameTooltip:ClearLines()
			local questId = self:BreadcrumbQuestId()
			local breadcrumbs = Grail:AvailableBreadcrumbs(questId)
			if nil ~= breadcrumbs then
				GameTooltip:AddLine(self.s['BREADCRUMB'])
				for i = 1, #breadcrumbs do
					GameTooltip:AddLine(self:PrettyQuestString({ breadcrumbs[i], self:ClassificationOfQuest(breadcrumbs[i]) }))
				end
				GameTooltip:Show()
			end
		end,

		BreadcrumbQuestId = function(self)
			local questId = GetQuestID()
			local questName = GetTitleText()

			-- Check the make sure the questId we are attempting to use makes sense with the title, otherwise
			-- the questId is incorrect and we need to try to get it
			if questName ~= self:QuestName(questId) then
				--	Attempt to get the information from the special quests since that information is cached.
				--	TODO: This code is taken from the internals of Grail and really should not be here.  Grail needs to provide this information.
				local foundQuest = nil
				local npcId = Grail:GetNPCId(questId, false)
				local questGivers = Grail.specialQuests[questName]
				if nil ~= questGivers then
					for i = 1, #questGivers do
						if tonumber(questGivers[i][1]) == npcId then
							foundQuest = questGivers[i][2]
						end
					end
				end
				if nil ~= foundQuest then
					questId = foundQuest
				else
					--	Look up the quest by name and use that questId
					questId = Grail:QuestWithName(questName)
				end
			end
			return questId
		end,

		ButtonEnter = function(self, button, ...)
			local aliasQuestId = Grail:AliasQuestId(button.questId)
			local questIdToUse = aliasQuestId or button.questId
			self:PopulateTooltipForQuest(button, questIdToUse, (questIdToUse ~= button.questId) and button.questId or nil)

			if not button.secureProcessed and not InCombatLockdown() then
				button:SetAttribute("type1", "click")
				button:SetAttribute("clickbutton", Wholly)
				button:SetAttribute("type2", "click")
				button:SetAttribute("alt-type1", "macro")
				button.secureProcessed = true
			else
				-- TODO: Should attempt a delayed setting of this if not button.secureProcessed and InCombatLockdown()
			end

			if 'P' == button.statusCode then
				self.lastPrerequisites = {}
				local lastIndexUsed = 0
				lastIndexUsed = self:PreparePrerequisiteInfo(Grail:QuestPrerequisitesSimple(button.questId), lastIndexUsed, true)
				lastIndexUsed = self:PreparePrerequisiteInfo(Grail:QuestPrerequisitesEver(button.questId), lastIndexUsed, true)
				lastIndexUsed = self:PreparePrerequisiteInfo(Grail:QuestPrerequisitesLog(button.questId), lastIndexUsed, true)
				lastIndexUsed = self:PreparePrerequisiteInfo(Grail:QuestPrerequisitesLogTurnin(button.questId), lastIndexUsed, true)
				lastIndexUsed = self:PreparePrerequisiteInfo(Grail:QuestPrerequisitesComplete(button.questId), lastIndexUsed, true)
				lastIndexUsed = self:PreparePrerequisiteInfo(Grail:QuestPrerequisitesCompleteTurnin(button.questId), lastIndexUsed, true)
				local outputString
				local started = false
				local tempTable = {}
				for questId, value in pairs(self.lastPrerequisites) do
					tinsert(tempTable, questId)
					outputString = ""
					if not started then
						self.tt:AddLine(" ")
						self.tt:AddLine("First in Prerequisite Chain:")
						started = true
					end
					for key, value2 in pairs(value) do
						if "" == outputString then
							outputString = "("..value2
						else
							outputString = outputString..","..value2
						end
					end
					outputString = outputString..") "
					self.tt:AddDoubleLine(outputString..self:PrettyQuestString({ questId, self:ClassificationOfQuest(questId) }), questId)
				end
				self.lastPrerequisites = started and tempTable or nil
			else
				self.lastPrerequisites = nil
			end

			if self.tt ~= GameTooltip then self.tt:Show() end
			GameTooltip:Show()
		end,

		ButtonPostClick = function(self, button)
			if button ~= self.clickingButton then print("Post click not from the same Pre click") end
			self.clickingButton = nil
		end,

		ButtonPreClick = function(self, button)
			self.clickingButton = button
		end,

		print_r = function(self, t, indent, done)
			done = done or {}
			indent = indent or ''
			local nextIndent
			for key, value in pairs (t) do
				if type (value) == "table" and not done[value] then
					nextIndent = nextIndent or (indent .. string.rep(' ',string.len(tostring (key))+2))
					done [value] = true
					print(indent .. "[" .. tostring (key) .. "] => Table {");
					print(nextIndent .. "{");
					self:print_r(value, nextIndent .. string.rep(' ',2), done)
					print(nextIndent .. "}")
				else
					print(indent .. "[" .. tostring (key) .. "] => " .. tostring (value).."")
				end
			end
		end,

		CallbackHandler = function(type, questId)
			Wholly:UpdateQuestCaches(true)
		end,

		CheckNPCTooltip = function(tooltip)
			if (not UnitIsPlayer("mouseover") or true) then
				-- check if this npc drops a quest item
				local id = Grail:GetNPCId(false, true)	-- only "mouseover" will be used
				local qs = id and Wholly.npcs[id] or nil
				if nil ~= qs then
					for _, questId in pairs(qs) do
						if Grail:CanAcceptQuest(questId) then
							local _, kindsOfNPC = Grail:IsTooltipNPC(id)
							if nil ~= kindsOfNPC then
								for i = 1, #(kindsOfNPC), 1 do
									local tooltipMessage = nil
									if kindsOfNPC[i][1] == Grail.NPC_TYPE_KILL then
										tooltipMessage = format(Wholly.s['KILL_TO_START_FORMAT'], Wholly:QuestName(questId))
									elseif kindsOfNPC[i][1] == Grail.NPC_TYPE_DROP then
										if Wholly:DroppedItemMatchesQuest(kindsOfNPC[i][2], questId) then
											tooltipMessage = format(Wholly.s['DROP_TO_START_FORMAT'], Grail:NPCName(kindsOfNPC[i][2]), Wholly:QuestName(questId))
										end
									end
									if nil ~= tooltipMessage then
										local leftStr = format("|TInterface\\MINIMAP\\ObjectIcons:17:9:0:0:128:128:40:56:32:63|t %s", tooltipMessage)
										tooltip:AddLine(leftStr);
									end
								end
							end
							tooltip:Show();
						end
					end
				end
			end
		end,

		ClassificationOfQuest = function(self, questId, shouldDisplayHolidays)
			local retval = 'U'
			local questBitMask
			if nil ~= questId and tonumber(questId) then
				questId = tonumber(questId)
				questBitMask = Grail:StatusCode(questId)
				if bit.band(questBitMask, Grail.bitMaskNonexistent + Grail.bitMaskError) > 0 then
					retval = 'U'
				elseif bit.band(questBitMask, Grail.bitMaskInLog) > 0 then
					retval = 'I'
				elseif bit.band(questBitMask, Grail.bitMaskQuestFailureWithAncestor - (shouldDisplayHolidays and Grail.bitMaskHoliday or 0) ) > 0 then
					retval = 'B'
				elseif bit.band(questBitMask, Grail.bitMaskCompleted) > 0 then
					retval = 'C'
				elseif bit.band(questBitMask, Grail.bitMaskPrerequisites) > 0 then
					retval = 'P'
				elseif Grail:IsDaily(questId) then	-- Grail.bitMaskResettable contains IsWeekly, IsMonthly and IsYearly, so we do not use because Blizzard shows yellow
					retval = 'D'
				elseif bit.band(questBitMask, Grail.bitMaskRepeatable) > 0 then
					retval = 'R'
				elseif bit.band(questBitMask, Grail.bitMaskLevelTooLow) > 0 then
					retval = 'L'
				elseif bit.band(questBitMask, Grail.bitMaskLowLevel) > 0 then
					retval = 'W'
				else
					retval = 'G'
				end
			end
			return retval
		end,

		---
		--	Gets all the quests in the map area, then classifies them based on the current player.
		ClassifyQuestsInMap = function(self, mapId)
			local retval = nil
			if nil ~= mapId and tonumber(mapId) then
				mapId = tonumber(mapId)
				local displaysHolidayQuestsAlways = false
				local showsLoremasterOnly = WhollyDatabase.showsLoremasterOnly
				if mapId >= Grail.mapAreaBaseHoliday and mapId <= Grail.mapAreaMaximumHoliday and WhollyDatabase.displaysHolidaysAlways then displaysHolidayQuestsAlways = true end
				retval = {}
				local questsInMap = Grail:QuestsInMap(mapId, WhollyDatabase.displaysDungeonQuests, showsLoremasterOnly) or {}
				for _,questId in pairs(questsInMap) do
					tinsert(retval, { questId, self:ClassificationOfQuest(questId, displaysHolidayQuestsAlways) })
				end
			end
			return retval
		end,

		-- This is named this way with this function signature because it is called from the SecureActionButtonTemplate exactly like this.
		Click = function(self, leftOrRight)
			if not TomTom or not TomTom.AddMFWaypoint then return end
			local button = self.clickingButton
			local questsToUse = {button.questId}
			local npcType = 'A'
			if "RightButton" == leftOrRight then
				if nil ~= self.lastPrerequisites then
					questsToUse = self.lastPrerequisites
				else
					npcType = 'T'
				end
			end
			self:AddDirectionalArrows(questsToUse, npcType)
		end,

		ConfigFrame_OnLoad = function(self, panel)
			panel.name = "Wholly"
			panel:Hide()
			InterfaceOptions_AddCategory(panel)
			local parent = panel:GetName()
			for i = 1, #self.configuration do
				local button = CreateFrame("CheckButton", parent.."Button"..i, panel, "InterfaceOptionsCheckButtonTemplate")
				button:SetPoint("TOPLEFT", panel, "TOPLEFT", 8, (i * -20) + 10)
				_G[button:GetName().."Text"]:SetText(self.configuration[i][0])
				button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:ScrollFrame_Update_WithCombatCheck() Wholly:UpdatePins() end)

				-- Special case for the button that matches the same functionality as the one on the world map
				if 12 == i then
					self.pairedConfigurationButton = button
					button:SetScript("OnClick", function(self)
						WhollyDatabase[Wholly.configuration[i][1]] = not WhollyDatabase[Wholly.configuration[i][1]]
						Wholly:UpdatePins()
						if Wholly.tooltip:IsVisible() and Wholly.tooltip:GetOwner() == Wholly.mapFrame then
							Wholly.tooltip:ClearLines()
							Wholly.tooltip:AddLine(Wholly.mapCountLine)
						end
					end)
				end

				-- Special case for the button that controls the display of the world map frame
				if 13 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:DisplayMapFrame(self:GetChecked()) end)
				end

				-- Special case for the button that controls whether quests in dungeons appear in the outer map
				-- This is done because the initial load of quests in an area which would normally get filtered by the default behavior needs to be reloaded
				if 14 == i or 19 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:UpdateQuestCaches(true) Wholly:ScrollFrame_Update_WithCombatCheck() Wholly:UpdatePins() end)
				end

				-- Special case for the breadcrumb control button
				if 15 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:UpdateBreadcrumb(Wholly) end)
				end

				-- Special case the holiday button
				if 16 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly.ForcePanelMapArea(Wholly) Wholly:ScrollFrame_Update_WithCombatCheck() end)
				end

				-- Special case the button that shows in log quest status
				if 18 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked()
									Wholly:ScrollFrame_Update_WithCombatCheck()
									end)
				end

				-- Special case the button that controls whether player coordinates are shown
				if 20 == i then
					self.pairedCoordinatesButton = button
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked()
									Wholly:UpdateCoordinateSystem()
									end)
				end

			end
		end,

		ConfigFrame_OnShow = function(self, panel)
			local parent = panel:GetName()
			for i = 1, #self.configuration do
				_G[parent.."Button"..i]:SetChecked(WhollyDatabase[self.configuration[i][1]])
			end
		end,

		DisplayMapFrame = function(self, shouldDisplay)
			if shouldDisplay then self.mapFrame:Show() else self.mapFrame:Hide() end
		end,

		Distance = function(self, x1, y1, x2, y2)
			local distRatio = WorldMapDetailFrame:GetHeight() / WorldMapDetailFrame:GetWidth();
			return math.sqrt( (x1 - x2)^2 + ((y1 - y2)/distRatio)^2 );
		end,

		Dropdown_AddButton = function(buttonTitle, level, hasArrow, mapId, anotherKey)
			local info = UIDropDownMenu_CreateInfo()
			info.hasArrow = hasArrow
			info.notCheckable = true
			info.text = buttonTitle
			local value = { ["Level1_Key"] = buttonTitle }
			if anotherKey then value["Level2_Key"] = anotherKey end
			info.value = value
			if not hasArrow then
				info.value = { ["mapID"] = mapId }
				info.func = function()
					Wholly.currentPanelMapID = mapId
					Wholly.ForcePanelMapArea(Wholly)
					CloseDropDownMenus()
				end
			end
			UIDropDownMenu_AddButton(info, level)
		end,

		---
		--	returns a table that is structured like the continents, sorted in localized name order
		Dropdown_Array = function(self, mapIdTable)
			local retval = {}
			for i = 1, #mapIdTable do
				tinsert(retval, { ["name"] = Grail.mapAreaMapping[mapIdTable[i]], ["mapID"] = mapIdTable[i] })
			end
			if #retval > 1 then
				table.sort(retval, function(a, b) return a["name"] < b["name"] end)
			end
			return retval
		end,

		Dropdown_ArrayReputation = function(self, reputationTable)
			local retval = {}
			local mapId
			for _, v in pairs(reputationTable) do
				mapId = v + Grail.mapAreaBaseReputationChange
				if nil ~= Grail.indexedQuests[mapId] and 0 < #(Grail.indexedQuests[mapId]) then
					tinsert(retval, { ["name"] = Grail.reputationMapping[v], ["mapID"] = mapId })
				end
			end
			if #retval > 1 then
				table.sort(retval, function(a, b) return a["name"] < b["name"] end)
			end
			return retval
		end,

		Dropdown_Initialize = function(self, level)
			level = level or 1
			local zoneCount
			local indexToUse
			if 1 == level then
				for i = 1, #(Grail.continents) do
					Wholly.Dropdown_AddButton(Grail.continents[i]["name"], level, true)
					zoneCount = #(Grail.continents[i]["zones"]) 
					if 4 == zoneCount then indexToUse = 5
					elseif 28 == zoneCount then indexToUse = 1
					elseif 39 == zoneCount then indexToUse = 2
					elseif 8 == zoneCount then indexToUse = 3
					elseif 12 == zoneCount then indexToUse = 4
					end
					Wholly.dungeonTest[indexToUse] = Grail.continents[i]["name"]
				end

				-- Dungeons
				Wholly.Dropdown_AddButton(BUG_CATEGORY3, level, true)

				-- World Events
				Wholly.Dropdown_AddButton(Wholly.s['WORLD_EVENTS'], level, true)

				-- Class
				Wholly.Dropdown_AddButton(CLASS, level, true)

				-- Professions
				Wholly.Dropdown_AddButton(TRADE_SKILLS, level, true)

				-- Reputation
				Wholly.Dropdown_AddButton(REPUTATION, level, true)

				-- Achievements
				Wholly.Dropdown_AddButton(ACHIEVEMENTS, level, true)

				-- Reputation Changes
				Wholly.Dropdown_AddButton(COMBAT_TEXT_SHOW_REPUTATION_TEXT, level, true)

				-- Daily
				Wholly.Dropdown_AddButton(DAILY, level, false, Grail.mapAreaBaseDaily)

				-- Other
				Wholly.Dropdown_AddButton(Wholly.s['OTHER'], level, true)

				-- Search
				Wholly.Dropdown_AddButton(SEARCH, level, true)

			elseif 3 == level then
				local Level1_Key = UIDROPDOWNMENU_MENU_VALUE["Level1_Key"]
				local Level2_Key = UIDROPDOWNMENU_MENU_VALUE["Level2_Key"]
				if Level1_Key == Wholly.dungeonTest[1] then
					if Level2_Key == BUG_CATEGORY3 then
						zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonKalimdorMapping)
					else
						zones = Wholly.Dropdown_Array(Wholly, Grail.achievements[Grail.playerFaction][1])
					end
				end
				if Level1_Key == Wholly.dungeonTest[2] then
					if Level2_Key == BUG_CATEGORY3 then
						zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonEasternKingdomsMapping)
					else
						zones = Wholly.Dropdown_Array(Wholly, Grail.achievements[Grail.playerFaction][2])
					end
				end
				if Level1_Key == Wholly.dungeonTest[3] then
					if Level2_Key == BUG_CATEGORY3 then
						zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonOutlandMapping)
					else
						zones = Wholly.Dropdown_Array(Wholly, Grail.achievements[Grail.playerFaction][3])
					end
				end
				if Level1_Key == Wholly.dungeonTest[4] then
					if Level2_Key == BUG_CATEGORY3 then
						zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonNorthrendMapping)
					else
						zones = Wholly.Dropdown_Array(Wholly, Grail.achievements[Grail.playerFaction][4])
					end
				end
				if Level1_Key == Wholly.dungeonTest[5] then
					if Level2_Key == BUG_CATEGORY3 then
						zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonMaelstromMapping)
					else
						zones = Wholly.Dropdown_Array(Wholly, Grail.achievements[Grail.playerFaction][5])
					end
				end
				if Level1_Key == EXPANSION_NAME0 then
					zones = Wholly.Dropdown_ArrayReputation(Wholly, Grail.reputationClassicMapping)
				end
				if Level1_Key == EXPANSION_NAME1 then
					zones = Wholly.Dropdown_ArrayReputation(Wholly, Grail.reputationBurningCrusadeMapping)
				end
				if Level1_Key == EXPANSION_NAME2 then
					zones = Wholly.Dropdown_ArrayReputation(Wholly, Grail.reputationWrathMapping)
				end
				if Level1_Key == EXPANSION_NAME3 then
					zones = Wholly.Dropdown_ArrayReputation(Wholly, Grail.reputationCataclysmMapping)
				end
				if Level2_Key == "World Event Achievements" then
					zones = Wholly.Dropdown_Array(Wholly, Grail.worldEventAchievements[Grail.playerFaction][Grail.reverseHolidayMapping[Level1_Key]])
				end
				for i = 1, #zones do
					Wholly.Dropdown_AddButton(zones[i]["name"], level, false, zones[i]["mapID"])
				end
			elseif 2 == level then
				local Level1_Key = UIDROPDOWNMENU_MENU_VALUE["Level1_Key"]
				if Level1_Key == BUG_CATEGORY3 then
					for i = 1, #(Grail.continents) do
						Wholly.Dropdown_AddButton(Grail.continents[i]["name"], level, true, nil, BUG_CATEGORY3)
					end
				end
				if Level1_Key == ACHIEVEMENTS then
					for i = 1, #(Grail.continents) do
						Wholly.Dropdown_AddButton(Grail.continents[i]["name"], level, true, nil, ACHIEVEMENTS)
					end
					for holidayKey, _ in pairs(Grail.worldEventAchievements[Grail.playerFaction]) do
						Wholly.Dropdown_AddButton(Grail.holidayMapping[holidayKey], level, true, nil, "World Event Achievements")
					end
					-- 5 Dungeon Achievement
					local achievementId = 4956
					Wholly.Dropdown_AddButton(select(2, GetAchievementInfo(achievementId)), level, false, Grail.mapAreaBaseAchievement + achievementId)
				end
				if Level1_Key == COMBAT_TEXT_SHOW_REPUTATION_TEXT then
					Wholly.Dropdown_AddButton(EXPANSION_NAME0, level, true)
					Wholly.Dropdown_AddButton(EXPANSION_NAME1, level, true)
					Wholly.Dropdown_AddButton(EXPANSION_NAME2, level, true)
					Wholly.Dropdown_AddButton(EXPANSION_NAME3, level, true)
				end
				local zones = {}
				for i = 1, #(Grail.continents) do
					if Level1_Key == Grail.continents[i]["name"] then
						zones = Grail.continents[i]["zones"]
					end
				end
				if Level1_Key == Wholly.s['WORLD_EVENTS'] then
					for code, name in pairs(Grail.holidayMapping) do
						tinsert(zones, { ["name"] = name, ["mapID"] = Grail.holidayToMapAreaMapping['H'..code] })
					end
					table.sort(zones, function(a, b) return a["name"] < b["name"] end)
				end
				if Level1_Key == CLASS then
					for code, englishName in pairs(Grail.classMapping) do
						local localizedGenderClassName = Grail:CreateClassNameLocalizedGenderized(englishName)
						local classColor = RAID_CLASS_COLORS[englishName]
						local mapId = Grail.classToMapAreaMapping['C'..code]
						if nil == classColor then
							classColor = { r = 1.0, g = 1.0, b = 1.0 }	-- TODO:	See if the actual color can be used instead of white
							localizedGenderClassName = "Monk"
						end	-- need to do for Monk currently
						if nil ~= Grail.mapAreaMapping[mapId] then
							tinsert(zones, { ["name"] = format("|cff%.2x%.2x%.2x%s|r", classColor.r*255, classColor.g*255, classColor.b*255, localizedGenderClassName), ["mapID"] = mapId, ["nonColorName"] = localizedGenderClassName })
						end
					end
					table.sort(zones, function(a, b) return a["nonColorName"] < b["nonColorName"] end)
				end
				if Level1_Key == TRADE_SKILLS then
					for code, professionName in pairs(Grail.professionMapping) do
						local mapId = Grail.professionToMapAreaMapping['P'..code]
						if nil ~= Grail.mapAreaMapping[mapId] then
							tinsert(zones, { ["name"] = professionName, ["mapID"] = mapId })
						end
					end
					table.sort(zones, function(a, b) return a["name"] < b["name"] end)
				end
				if Level1_Key == REPUTATION then
					for reputationIndex, reputationName in pairs(Grail.reputationMapping) do
						local mapId = Grail.mapAreaBaseReputation + reputationIndex
						if nil ~= Grail.mapAreaMapping[mapId] then
							tinsert(zones, { ["name"] = reputationName, ["mapID"] = mapId })
						end
					end
					table.sort(zones, function(a, b) return a["name"] < b["name"] end)
				end
				if Level1_Key == Wholly.s['OTHER'] then
					zones = Wholly.Dropdown_Array(Wholly, Grail.otherMapping)
				end
				if Level1_Key == SEARCH then
					local info = UIDropDownMenu_CreateInfo()
					info.hasArrow = false
					info.notCheckable = true
					info.text = Wholly.s['SEARCH_NEW']
					info.func = function()
						com_mithrandir_whollySearchFrame:Show()
						CloseDropDownMenus()
					end
					UIDropDownMenu_AddButton(info, level)
					if WhollyDatabase.searches and 0 < #(WhollyDatabase.searches) then
						for i = 1, #(WhollyDatabase.searches) do
							info = UIDropDownMenu_CreateInfo()
							info.hasArrow = false
							info.notCheckable = true
							info.text = SEARCH .. ': ' .. WhollyDatabase.searches[i]
							info.value = { ["mapID"] = 0 }
							info.func = function()
								Wholly.SearchForQuestNamesMatching(Wholly, WhollyDatabase.searches[i])
								Wholly.currentPanelMapID = 0
								Wholly.ForcePanelMapArea(Wholly)
								CloseDropDownMenus()
							end
							UIDropDownMenu_AddButton(info, level)
						end
						-- now add the Clear menu item
						info = UIDropDownMenu_CreateInfo()
						info.hasArrow = false
						info.notCheckable = true
						info.text = Wholly.s['SEARCH_CLEAR']
						info.func = function()
							WhollyDatabase.searches = nil
							CloseDropDownMenus()
						end
						UIDropDownMenu_AddButton(info, level)
					end
					zones = nil
				end
				if nil ~= zones then
					for i = 1, #zones do
						Wholly.Dropdown_AddButton(zones[i]["name"], level, false, zones[i]["mapID"])
					end
				end
			end
		end,

		DroppedItemMatchesQuest = function(self, dropNPCCode, matchingQuestId)
			local retval = true
			local questCodes = Grail:CodesWithPrefixNPC(dropNPCCode, "Q:")
			if nil ~= questCodes then
				retval = false
				for i = 1, #questCodes do
					local quests = { strsplit(",", strsub(questCodes[i], 3)) }
					for j = 1, #quests do
						if tonumber(quests[j]) == tonumber(matchingQuestId) then retval = true end
					end
				end
			end
			return retval
		end,

		FilterQuests = function(self, forPanel)
			local f = forPanel and self.filteredPanelQuests or self.filteredPinQuests
			f = {}
			local questsInMap = forPanel and self.cachedPanelQuests or self.cachedPinQuests
			local shouldAdd, statusCode

			--	We want to be able to force display of quests that are class or profession specific
			--	unless they are associated with the player.  In that case, the display of the quests
			--	obeys the same rules as the quests in a normal map area.
			local shouldForce = false
			if nil ~= self.currentPanelMapID and self.currentPanelMapID >= Grail.mapAreaBaseClass and self.currentPanelMapID <= Grail.mapAreaMaximumProfession then
				shouldForce = true
				if self.playerClassMap == self.currentPanelMapID then shouldForce = false end
				if self.currentPanelMapID >= Grail.mapAreaBaseProfession then
					for key,value in pairs(Grail.professionToMapAreaMapping) do
						if value == self.currentPanelMapID then
							local actualKey = key:sub(2, 2)
							if Grail:ProfessionExceeds(actualKey, 1) then -- indicates the profession is known
								shouldForce = false
							end
						end
					end
				end
			end
			local repuationQuest = false
			if nil ~= self.currentPanelMapID and self.currentPanelMapID > Grail.mapAreaBaseReputation and self.currentPanelMapID <= Grail.mapAreaMaximumReputation then
				reputationQuest = true
			end

			for i = 1, #questsInMap do
				statusCode = questsInMap[i][2]

				if 'W' == statusCode then shouldAdd = WhollyDatabase.showsLowLevelQuests
				elseif 'I' == statusCode then shouldAdd = forPanel and WhollyDatabase.showsQuestsInLog
				elseif 'C' == statusCode then shouldAdd = forPanel and WhollyDatabase.showsCompletedQuests
				elseif 'B' == statusCode then
					local reputationForce = false
					if reputationQuest then
						local status = Grail:StatusCode(questsInMap[i][1])
						local bitResult = bit.band(status, Grail.bitMaskQuestFailureWithAncestor - (Grail.bitMaskAncestorReputation + Grail.bitMaskReputation))
						if bitResult == 0 then		-- the quest only failed because of reputation failure
							reputationForce = true
						end
					end
					shouldAdd = forPanel and (WhollyDatabase.showsUnobtainableQuests or shouldForce or reputationForce)
				elseif 'D' == statusCode then shouldAdd = WhollyDatabase.showsDailyQuests
				elseif 'L' == statusCode then shouldAdd = WhollyDatabase.showsHighLevelQuests
				elseif 'P' == statusCode then shouldAdd = WhollyDatabase.showsQuestsThatFailPrerequsites
				else shouldAdd = true
				end

				if shouldAdd then
					tinsert(f, questsInMap[i])
				end
			end
			if forPanel then self.filteredPanelQuests = f else self.filteredPinQuests = f end
			if not forPanel then
				self.mapCountLine = self:PrettyQuestCountString(questsInMap, #(self.filteredPinQuests), true)
			else
				self.panelCountLine = self:PrettyQuestCountString(questsInMap, #(self.filteredPanelQuests))
			end
		end,

		FilterPanelQuests = function(self)
			self:FilterQuests(true)
		end,

		FilterPinQuests = function(self)
			self:FilterQuests(false)
		end,

		ForcePanelMapArea = function(self)
			local mapAreaName = Grail.mapAreaMapping[self.currentPanelMapID] or GetRealZoneText()	-- default to something if we do not support the zone
			if nil ~= mapAreaName then UIDropDownMenu_SetText(self.dropdown, mapAreaName) end
			self.cachedPanelQuests = self:ClassifyQuestsInMap(self.currentPanelMapID) or {}
			self:ScrollFrame_Update_WithCombatCheck()
		end,

		GetMousePosition = function(self)
			local left, top = WorldMapDetailFrame:GetLeft(), WorldMapDetailFrame:GetTop();
			local width, height = WorldMapDetailFrame:GetWidth(), WorldMapDetailFrame:GetHeight();
			local scale = WorldMapDetailFrame:GetEffectiveScale();

			local x, y = GetCursorPosition();
			local cx = (x/scale - left) / width;
			local cy = (top - y/scale) / height;
	
			return math.min(math.max(cx, 0), 1), math.min(math.max(cy, 0), 1);
		end,

		GetPin = function(self, npcId)
			if nil ~= self.regNPCPins[npcId] then return self.regNPCPins[npcId] end		-- already have a pin for this npcId

			if #self.regFrames > 0 then
				local pin = table.remove(self.regFrames)
				pin.npcId = npcId
			end

			local pin = CreateFrame("Frame", nil, WorldMapDetailFrame);
			pin.npcId = npcId
			pin:SetWidth(16);
			pin:SetHeight(16);
			pin:EnableMouse(true);
			pin:SetScript("OnEnter", function(pin) self:ShowTooltip(pin) end)
			pin:SetScript("OnLeave", function() self:HideTooltip() end)
			pin.SetType = function(self, texType)
				if self.texType == texType then return end -- don't need to make changes
				local width, height = 0.125, 0.125

				-- 256x64, 8x4
				self.texture:SetDesaturated(false);
				self.texture:SetVertexColor(1, 1, 1);
				if texType == "D" then
					self.texture:SetTexCoord(3*width, 4*width, 1*height, 2*height);
				elseif texType == "R" then
					self.texture:SetTexCoord(4*width, 5*width, 1*height, 2*height);
				elseif texType == "P" then
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
					self.texture:SetVertexColor(1.0, 0.0, 0.0);
				elseif texType == "W" then
					-- I don't like this one, it's too hard to see.
					-- self.texture:SetTexCoord(4*width, 5*width, 3*height, 4*height); -- this is the actual dimmed quest giver icon
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
					self.texture:SetVertexColor(0.75, 0.75, 0.75);
				elseif texType == "L" then
					-- Note: not all systems can desaturate. If they don't, I don't really care.
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
					self.texture:SetDesaturated(1);
				else
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
				end

				self.texType = texType
			end

			pin.texture = pin:CreateTexture();
			pin.texture:SetTexture("Interface\\MINIMAP\\ObjectIcons.blp");
			pin.texture:SetAllPoints();
			pin:SetType("L")
			self.regNPCPins[npcId] = pin
			return pin;
		end,

		GetPrerequisiteInfo = function(self, questId, index)
			local statusLetter = self:ClassificationOfQuest(questId)
			if 'P' == statusLetter then
				self:PreparePrerequisiteInfo(Grail:QuestPrerequisites(questId), index, false)
			elseif 'U' == statusLetter or 'B' == statusLetter or 'C' == statusLetter then -- or 'L' == statusLetter
				-- do nothing since this is a failure
			else	-- I, D, R, G, W
				if nil == self.lastPrerequisites[questId] then self.lastPrerequisites[questId] = {} end
				if not tContains(self.lastPrerequisites[questId], index) then tinsert(self.lastPrerequisites[questId], index) end
			end
		end,

		HideAllPins = function(self)
			for i, v in pairs(self.regPins) do
				self:HidePin(i)
			end
		end,

		HidePin = function(self, id)
			if self.regPins[id] then
				self.regPins[id]:Hide()
				table.insert(self.regFrames, self.regPins[id])
				self.regNPCPins[self.regPins[id].npcId] = nil
				self.regPins[id] = nil
			end
		end,

		HideTooltip = function(self)
			self.tooltip:Hide()
		end,

		LoadDefaults = function(self)
			local db = {}
			db.defaultsLoaded = true
			db.prependsQuestLevel = true
			db.appendRequiredLevel = true
			db.showsLowLevelQuests = true
			db.showsAnyPreviousRepeatableCompletions = true
			db.updatesPanelWhenZoneChanges = true
			db.displaysMapPins = true
			db.displaysMapFrame = true
			db.displaysDungeonQuests = true
			db.displaysBreadcrumbs = true
			db.displaysHolidaysAlways = true
			db.updatesWorldMapOnZoneChange = true
			db.showsInLogQuestStatus = true
			db.version = Wholly.versionNumber
			WhollyDatabase = db
		end,

		NPCInfoSection = function(self, heading, table, button, meetsCriteria)
			if nil == table then return end
			self.tt:AddLine(" ")
			self.tt:AddLine(heading)
			for _, npcId in pairs(table) do
				local locations, npcName = Grail:LocationNPCInMap(npcId)
				if nil ~= locations then
					for _, location in pairs(locations) do
						local realAreaId = nil
						local locationString
						if 1 == #location then
							locationString = location[1]
							if strfind(locationString, 'Near') and 4 < strlen(locationString) then
								locationString = "Near "..Grail.mapAreaMapping[tonumber(strsub(locationString, 5))]
							elseif nil ~= tonumber(locationString) then
								locationString = Grail.mapAreaMapping[tonumber(locationString)]
							end
						else
							if 4 == #location then
								realAreaId = tonumber(location[4])
							end
							local mapAreaName = Grail.mapAreaMapping[tonumber(location[1])]
							if nil == mapAreaName then mapAreaName = "NO MAP" end
							local coords = location[2]
							if nil == coords then coords = "NO COORDS" end
							locationString = mapAreaName .. " " .. coords

							if meetsCriteria then
								local desiredMacroValue = SLASH_TARGET1 .. " " .. npcName
								if button:GetAttribute("macrotext") ~= desiredMacroValue and not InCombatLockdown() then
									button:SetAttribute("macrotext", desiredMacroValue)
								end
							end

						end
						self.tt:AddDoubleLine(self:PrettyNPCString(npcName, false, realAreaId), locationString)
					end
				end
			end
		end,

		OnEvent = function(self, frame, event, ...)
			if self.eventDispatch[event] then
				self.eventDispatch[event](self, frame, ...)
			end
		end,

		OnHide = function(self, frame)
		end,

		OnLoad = function(self, frame)
			if not Grail or Grail.versionNumber < requiredGrailVersion then print(format(self.s['REQUIRES_FORMAT'], requiredGrailVersion)) return end
			SlashCmdList["WHOLLY"] = function(msg)
				self:SlashCommand(frame, msg)
			end
			SLASH_WHOLLY1 = "/wholly"
			com_mithrandir_whollyFrameTitleText:SetText("Wholly ".. com_mithrandir_whollyFrameTitleText:GetText())

			-- The frame is not allowing button presses to things just on the outside of its bounds so we move the hit rect
			frame:SetHitRectInsets(0, 32, 0, 84)

			local LibStub = _G["LibStub"]
			if LibStub then
				local LDB = LibStub("LibDataBroker-1.1", true)
				if LDB then
					local launcher = LDB:NewDataObject("Wholly", { type="launcher", icon="Interface\\Icons\\INV_Misc_Book_07",
							OnClick = function(theFrame, button) if button == "RightButton" then InterfaceOptionsFrame_OpenToCategory("Wholly") else frame:Show() end end,
							OnTooltipShow = function(tooltip) Wholly:ProcessInitialUpdate() Wholly.ldbTooltipOwner = tooltip:GetOwner()
							tooltip:AddLine("Wholly - " .. UIDropDownMenu_GetText(Wholly.dropdown) ) tooltip:AddLine(Wholly.panelCountLine) end, 
							})
					self.coordinates = LDB:NewDataObject("Wholly Coordinates", { type="data source", icon="Interface\\Icons\\INV_Misc_Map02", text="",
							OnClick = function(theFrame, button) Wholly.pairedCoordinatesButton:Click() end,
							OnTooltipShow = function(tooltip)
								Wholly.ldbCoordinatesTooltipOwner = tooltip:GetOwner()
								local dungeonLevel = GetCurrentMapDungeonLevel()
								local dungeonIndicator = (dungeonLevel > 0) and "["..dungeonLevel.."]" or ""
								local mapAreaId = GetCurrentMapAreaID()
								local mapAreaName = Grail.mapAreaMapping[mapAreaId] or "UNKNOWN"
								tooltip:AddLine(string.format("%d%s %s", mapAreaId, dungeonIndicator, mapAreaName)) end,
							})
				end
			end

			self.dropdown = CreateFrame("Button", frame:GetName().."ZoneButton", frame, "UIDropDownMenuTemplate")
			UIDropDownMenu_Initialize(self.dropdown, self.Dropdown_Initialize) -- took away "MENU" because no show with it
			self.dropdown:SetPoint("TOPLEFT", frame, "TOPLEFT", 60, -40)
			UIDropDownMenu_SetWidth(self.dropdown, 240, 0)
			UIDropDownMenu_JustifyText(self.dropdown, "LEFT")
			-- By default, the dropdown has it clicking work with the little button on the right.  This makes it work for the whole button:
			self.dropdown:SetScript("OnClick", function(self) ToggleDropDownMenu(nil, nil, Wholly.dropdown) PlaySound("igMainMenuOptionCheckBoxOn") end)

			self.tooltip = CreateFrame("GameTooltip", "com_mithrandir_WhollyTooltip", UIParent, "GameTooltipTemplate");
			self.tooltip:SetFrameStrata("TOOLTIP");
			self.tooltip.large = com_mithrandir_WhollyTooltipTextLeft1:GetFontObject();
			self.tooltip.small = com_mithrandir_WhollyTooltipTextLeft2:GetFontObject();
			self.tooltip.SetLastFont = function(self, fontObj, rightText)
				local txt = rightText and "Right" or "Left"
				_G[format("com_mithrandir_WhollyTooltipText%s%d", txt, self:NumLines())]:SetFont(fontObj:GetFont())
			end

			self.tt2 = CreateFrame("GameTooltip", "com_mithrandir_WhollyOtherTooltip", GameTooltip, "GameTooltipTemplate")

			local f = CreateFrame("Button", nil, WorldMapFrame, "UIPanelButtonTemplate")
			f:SetSize(100, 25)
			if nil == Gatherer_WorldMapDisplay then
				f:SetPoint("TOPLEFT", WorldMapPositioningGuide, "TOPLEFT", 4, -4)
			else
				f:SetPoint("TOPLEFT", Gatherer_WorldMapDisplay, "TOPRIGHT", 4, 0)
			end
			f:SetFrameStrata("HIGH")	-- if not set like this, will not respond when the map is small
			f:SetScale(0.7)
			f:SetText("Wholly")
			f:SetScript("OnShow", function(self)
									if nil == Gatherer_WorldMapDisplay then
										if not(GetCVarBool("miniWorldMap")) then
											self:SetPoint("TOPLEFT", WorldMapPositioningGuide, "TOPLEFT", 4, -4)
										else
											self:SetPoint("TOPLEFT", WorldMapTitleButton, "TOPLEFT", 8, -3)
										end
									else
										self:SetPoint("TOPLEFT", Gatherer_WorldMapDisplay, "TOPRIGHT", 4, 0)
									end
								end)
			f:SetScript("OnEnter", function(self) local t = Wholly.tooltip t:ClearLines() t:SetOwner(self) t:AddLine(Wholly.mapCountLine) t:Show() t:ClearAllPoints() t:SetPoint("TOPLEFT", self, "BOTTOMRIGHT") end)
			f:SetScript("OnLeave", function(self) Wholly.tooltip:Hide() end)
			f:SetScript("OnClick", function(self) Wholly.pairedConfigurationButton:Click() end)
			f:Hide()
			self.mapFrame = f

			-- if the UI panel disappears (maximized WorldMapFrame) we need to change parents
			UIParent:HookScript("OnHide", function()
				self.tooltip:SetParent(WorldMapFrame);
				self.tooltip:SetFrameStrata("TOOLTIP");
			end)
			UIParent:HookScript("OnShow", function()
				self.tooltip:SetParent(UIParent);
				self.tooltip:SetFrameStrata("TOOLTIP");
			end)

			GameTooltip:HookScript("OnTooltipSetUnit", Wholly.CheckNPCTooltip)

			local nf = CreateFrame("Frame")
			nf:SetScript("OnEvent", function(frame, event, ...) self:OnEvent(frame, event, ...) end)
			nf:RegisterEvent("ADDON_LOADED")
			self.notificationFrame = nf

		end,

		---
		--	The first time the panel is shown it is populated with the information from the current map area.
		OnShow = function(self, frame)
			self:ProcessInitialUpdate()
			if WhollyDatabase.showsInLogQuestStatus then
				self:ScrollFrame_Update_WithCombatCheck()
			end
		end,

		OnUpdate = function(self, frame, elapsed)
			self.lastUpdate = self.lastUpdate + elapsed
			if self.lastUpdate < self.updateThreshold then return end
			local x, y = GetPlayerMapPosition('player')
			if self.previousX ~= x or self.previousY ~= y then
				if nil ~= self.coordinates then
					self.coordinates.text = string.format("%.2f, %.2f", x * 100, y * 100)
				end
				self.previousX = x
				self.previousY = y
			end
			self.lastUpdate = 0
		end,

		PrettyNPCString = function(self, npcName, mustKill, realAreaId)
			if mustKill then npcName = format(self.s['MUST_KILL_PIN_FORMAT'], npcName) end
			if realAreaId then npcName = format("%s => %s", npcName, Grail.mapAreaMapping[realAreaId]) end
			return npcName
		end,

		PrettyQuestCountString = function(self, questTable, displayedCount, forMap)
			local retval = ""
			displayedCount = displayedCount or 0
			if nil ~= questTable then
				local totals = { ['B'] = 0, ['C'] = 0, ['D'] = 0, ['G'] = 0, ['I'] = 0, ['L'] = 0, ['P'] = 0, ['R'] = 0, ['U'] = 0, ['W'] = 0, }
				local code
				for i = 1, #questTable do
					code = questTable[i][2]
					totals[code] = totals[code] + 1
				end
				local colorCode
				for _,code in pairs({ 'G', 'W', 'L', 'P', 'D', 'R', 'I', 'C', 'B', }) do
					colorCode = self.color[code]
					retval = retval .. "|c" .. colorCode .. totals[code] .. "|r/"
				end
				local displayStart, displayEnd = "", ""
				if forMap and not WhollyDatabase.displaysMapPins then
					displayStart, displayEnd = "|cffff0000", "|r"
				end
				retval = retval .. "|c" .. self.color['U'] .. totals['U'] .. "|r" .. "  [" .. displayStart .. displayedCount .. displayEnd .. "/" .. #questTable .."]"
			end
			return retval
		end,

		PrettyQuestString = function(self, questTable)
			local questId = questTable[1]
			local filterCode = questTable[2]
			local statusCode = Grail:StatusCode(questId)
			local questLevel = Grail:QuestLevel(questId)
			local questLevelString = WhollyDatabase.prependsQuestLevel and format("[%s] ", questLevel or "??") or ""
			local requiredLevelString = ""
			if WhollyDatabase.appendRequiredLevel then
				local success, _, questLevelNeeded, _ = Grail:MeetsRequirementLevel(questId)
				if bit.band(statusCode, Grail.bitMaskLevelTooLow) > 0 then requiredLevelString = format(" [%s]", questLevelNeeded) end
			end
			local colorCode = self.color[filterCode]
--			local repeatableCompletedString = ('D' == filterCode or 'R' == filterCode) and WhollyDatabase.showsAnyPreviousRepeatableCompletions and bit.band(statusCode, Grail.bitMaskEverCompleted) > 0 and "*" or ""
			local repeatableCompletedString = WhollyDatabase.showsAnyPreviousRepeatableCompletions and bit.band(statusCode, Grail.bitMaskResettableRepeatableCompleted) > 0 and "*" or ""
			return format("|c%s%s%s%s%s|r", colorCode, questLevelString, self:QuestName(questId), repeatableCompletedString, requiredLevelString)
		end,

		ProcessInitialUpdate = function(self)
			if not self.initialUpdateProcessed then
				self.currentPanelMapID = GetCurrentMapAreaID()
				self:ForcePanelMapArea()
				self.initialUpdateProcessed = true
			end
		end,
			
		PopulateTooltipForQuest = function(self, frame, questId, aliasQuestId)
			self.tt = GameTooltip
			questId = tonumber(questId)
			self.tt:SetOwner(frame, "ANCHOR_RIGHT")
			self.tt:ClearLines()
			if nil == questId then return end
			self.tt:SetHyperlink(format("quest:%d", questId))
			if not Grail:DoesQuestExist(questId) then self.tt:AddLine(" ") self.tt:AddLine(self.s['GRAIL_NOT_HAVE']) return end

			if self.debug then
				self.tt:AddLine(" ")
				local aliasQuestString = aliasQuestId and " ("..aliasQuestId..")" or ""
				self.tt:AddLine(self.s['QUEST_ID']..questId..aliasQuestString)
			end

			questId = aliasQuestId or questId	-- remap to the alias now that the Blizzard interaction is done
			local obtainersCode = Grail:CodeObtainers(questId)
			local holidayCode = Grail:CodeHoliday(questId)
			local questLevel = Grail:QuestLevel(questId)
			local _, _, requiredLevel, notToExceedLevel = Grail:MeetsRequirementLevel(questId)
			local questType = self:QuestTypeString(questId)
			local statusCode = Grail:StatusCode(questId)
			local normalColor, redColor, orangeColor, greenColor = "ffffd200", "ffff0000", "ffff9900", "ff00ff00"
			local colorCode

			self.tt:AddLine(" ")
			self.tt:AddDoubleLine(LEVEL, questLevel)
			self.tt:AddDoubleLine(self.s['REQUIRED_LEVEL'], requiredLevel)
			if bit.band(statusCode, Grail.bitMaskLevelTooHigh) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorLevelTooHigh) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..GUILD_RECRUITMENT_MAXLEVEL.."|r", (127 == notToExceedLevel) and self.s['MAXIMUM_LEVEL_NONE'] or notToExceedLevel)

			if "" == questType then questType = self.s['QUEST_TYPE_NORMAL'] end
			local len = strlen(questType)
			if len > 0 and ' ' == strsub(questType, len, len) then questType = strsub(questType, 1, len - 1) end
			self.tt:AddDoubleLine(TYPE, questType)

			local loremasterString = self.s['MAPAREA_NONE']
			local loremasterMapArea = Grail:LoremasterMapArea(questId)
			if nil ~= loremasterMapArea then loremasterString = Grail.mapAreaMapping[loremasterMapArea] end
			self.tt:AddDoubleLine(self.s['LOREMASTER_AREA'], loremasterString)

			self.tt:AddLine(" ")
			local factionString = FACTION_OTHER
			if Grail.bitMaskFactionAll == bit.band(obtainersCode, Grail.bitMaskFactionAll) then
				factionString = self.s['FACTION_BOTH']
			elseif 0 < bit.band(obtainersCode, Grail.bitMaskFactionAlliance) then
				factionString = FACTION_ALLIANCE
			elseif 0 < bit.band(obtainersCode, Grail.bitMaskFactionHorde) then
				factionString = FACTION_HORDE
			end
			if bit.band(statusCode, Grail.bitMaskFaction) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorFaction) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..FACTION.."|r", factionString)

			local classString
			if 0 == bit.band(obtainersCode, Grail.bitMaskClassAll) then
				classString = self.s['CLASS_NONE']
			elseif Grail.bitMaskClassAll == bit.band(obtainersCode, Grail.bitMaskClassAll) then
				classString = self.s['CLASS_ANY']
			else
				classString = ""
				for letterCode, bitValue in pairs(Grail.classToBitMapping) do
					if 0 < bit.band(obtainersCode, bitValue) then
						local englishName = Grail.classMapping[letterCode]
						local localizedGenderClassName = Grail:CreateClassNameLocalizedGenderized(englishName)
						local classColor = RAID_CLASS_COLORS[englishName]
						-- TODO: Remove the following line when we actually have Monks available
						if nil == classColor then classColor = { r = 1.0, g = 1.0, b = 1.0 } localizedGenderClassName = "Monk" end
						classString = classString .. format("|cff%.2x%.2x%.2x%s|r ", classColor.r*255, classColor.g*255, classColor.b*255, localizedGenderClassName)
					end
				end
				len = strlen(classString)
				if 0 > len and ' ' == strsub(classString, len, len) then classString = strsub(classString, 1, len - 1) end
			end
			if bit.band(statusCode, Grail.bitMaskClass) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorClass) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..CLASS.."|r", classString)

			local genderString = self.s['GENDER_NONE']
			if Grail.bitMaskGenderAll == bit.band(obtainersCode, Grail.bitMaskGenderAll) then
				genderString = self.s['GENDER_BOTH']
			elseif 0 < bit.band(obtainersCode, Grail.bitMaskGenderMale) then
				genderString = MALE
			elseif 0 < bit.band(obtainersCode, Grail.bitMaskGenderFemale) then
				genderString = FEMALE
			end
			if bit.band(statusCode, Grail.bitMaskGender) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorGender) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..self.s['GENDER'].."|r", genderString)

			-- Note that race can show races of any faction, especially if the quest is marked just to exclude a specific race
			local raceString
			if 0 == bit.band(obtainersCode, Grail.bitMaskRaceAll) then
				raceString = self.s['RACE_NONE']
			elseif Grail.bitMaskRaceAll == bit.band(obtainersCode, Grail.bitMaskRaceAll) then
				raceString = self.s['RACE_ANY']
			else
				raceString = ""
				for letterCode, bitValue in pairs(Grail.raceToBitMapping) do
					if 0 < bit.band(obtainersCode, bitValue) then
						local englishName = Grail.raceMapping[letterCode]
						local localizedGenderRaceName = Grail:CreateRaceNameLocalizedGenderized(englishName)
						-- TODO: Remove the following line when we actually have Pandarens available
						if nil == localizedGenderRaceName then localizedGenderRaceName = "Pandaren" end
						raceString = raceString .. localizedGenderRaceName .. " "
					end
				end
				len = strlen(raceString)
				if 0 > len and ' ' == strsub(raceString, len, len) then raceString = strsub(raceString, 1, len - 1) end
			end
			if bit.band(statusCode, Grail.bitMaskRace) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorRace) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..RACES.."|r", raceString)

			if 0 ~= holidayCode then
				self.tt:AddLine(" ")
				if bit.band(statusCode, Grail.bitMaskHoliday) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorHoliday) > 0 then colorCode = orangeColor else colorCode = normalColor end
				self.tt:AddLine("|c"..colorCode..self.s['HOLIDAYS_ONLY'].."|r")
				for letterCode, bitValue in pairs(Grail.holidayToBitMapping) do
					if 0 < bit.band(holidayCode, bitValue) then
						self.tt:AddLine(Grail.holidayMapping[letterCode])
					end
				end
			end

			if Grail.quests[questId]['SP'] then
				self.tt:AddLine(" ")
				self.tt:AddLine(self.s['SP_MESSAGE'])
			end

			if nil ~= Grail.quests[questId]['rep'] then
				local factionStandingFormat = "FACTION_STANDING_LABEL%d"
				if Grail.playerGender == 3 then factionStandingFormat = factionStandingFormat.."_FEMALE" end
				self.tt:AddLine(" ")
				if bit.band(statusCode, Grail.bitMaskReputation) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorReputation) > 0 then colorCode = orangeColor else colorCode = normalColor end
				self.tt:AddLine("|c"..colorCode..self.s['REPUTATION_REQUIRED'].."|r")
				for reputationIndex, repTable in pairs(Grail.quests[questId]['rep']) do
					-- repTable can have 'min' and/or 'max'
					local reputationString = ""
					local repValue = repTable['min']
					if nil ~= repValue then
						reputationString = format("|cFF00FF00%s|r", GetText(format(factionStandingFormat, self.reputationLevelMapping[repValue])))
					end
					repValue = repTable['max']
					if nil ~= repValue then
						if 0 < strlen(reputationString) then reputationString = reputationString .. " - " end
						reputationString = reputationString .. format("|cFFFF0000%s|r", GetText(format(factionStandingFormat, self.reputationLevelMapping[repValue])))
					end
					self.tt:AddDoubleLine(Grail.reputationMapping[reputationIndex], reputationString)
				end
			end

			if nil ~= Grail.quests[questId]['prof'] then
				self.tt:AddLine(" ")
				if bit.band(statusCode, Grail.bitMaskProfession) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorProfession) > 0 then colorCode = orangeColor else colorCode = normalColor end
				self.tt:AddLine("|c"..colorCode..TUTORIAL_TITLE38..':'.."|r")
				for code, value in pairs(Grail.quests[questId]['prof']) do
					self.tt:AddDoubleLine(Grail.professionMapping[code], "|cFF00FF00"..value.."|r")
				end
			end

			self:QuestInfoSection(self.s['BREADCRUMB'], Grail:QuestBreadcrumbs(questId))
			if bit.band(statusCode, Grail.bitMaskInvalidated) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorInvalidated) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self:QuestInfoSection("|c"..colorCode..self.s['INVALIDATE'].."|r", Grail:QuestInvalidates(questId))
			local lastIndexUsed = 0
			lastIndexUsed = self:QuestInfoSection(self.s['PREREQUISITES'], Grail:QuestPrerequisitesSimple(questId), lastIndexUsed)
			lastIndexUsed = self:QuestInfoSection(self.s['PREQ_EVER'], Grail:QuestPrerequisitesEver(questId), lastIndexUsed)
			lastIndexUsed = self:QuestInfoSection(self.s['PREQ_LOG'], Grail:QuestPrerequisitesLog(questId), lastIndexUsed)
			lastIndexUsed = self:QuestInfoSection(self.s['PREQ_LOG_TURN'], Grail:QuestPrerequisitesLogTurnin(questId), lastIndexUsed)
			lastIndexUsed = self:QuestInfoSection(self.s['PREQ_COMPLETE'], Grail:QuestPrerequisitesComplete(questId), lastIndexUsed)
			lastIndexUsed = self:QuestInfoSection(self.s['PREQ_COMPLETE_TURN'], Grail:QuestPrerequisitesCompleteTurnin(questId), lastIndexUsed)

			if self.tt:NumLines() > 50 then
				self.tt2:SetOwner(GameTooltip, "ANCHOR_RIGHT")
				self.tt2:ClearLines()
				self.tt = self.tt2
			end

			self:QuestInfoSection(self.s['OAC'], Grail:QuestOnAcceptCompletes(questId))
			self:QuestInfoSection(self.s['OCC'], Grail:QuestOnCompletionCompletes(questId))
			self:QuestInfoTurninSection(self.s['OTC'], Grail:QuestOnTurninCompletes(questId))
			if nil ~= Grail.quests[questId]['AZ'] then
				self.tt:AddLine(" ")
				self.tt:AddDoubleLine(self.s['ENTER_ZONE'], Grail.mapAreaMapping[Grail.quests[questId]['AZ']])
			end

			if nil ~= Grail.quests[questId][6] and 0 < #(Grail.quests[questId][6]) then
				self.tt:AddLine(" ")
				self.tt:AddLine(COMBAT_TEXT_SHOW_REPUTATION_TEXT .. ':')
				local index, value
				for _, repString in pairs(Grail.quests[questId][6]) do
					index = tonumber(strsub(repString, 1, 2))
					value = tonumber(strsub(repString, 3))
					if value > 0 then
						colorCode = greenColor
					else
						colorCode = redColor
						value = -1 * value
					end
					self.tt:AddDoubleLine(Grail.reputationMapping[index], "|c"..colorCode..value.."|r")
				end
			end

			self:NPCInfoSection(self.s['WHEN_KILL'], Grail:QuestNPCKills(questId), frame, false)
			self:NPCInfoSection(TUTORIAL_TITLE1..':', Grail:QuestNPCAccepts(questId), frame, ('I' ~= frame.statusCode))
			self:NPCInfoSection(TURN_IN_QUEST..':', Grail:QuestNPCTurnins(questId), frame, ('I' == frame.statusCode))
		end,

		PreparePrerequisiteInfo = function(self, table, lastIndexUsed, doMath)
			if nil == table then return lastIndexUsed end
			local index = lastIndexUsed or 0
			for key, value in pairs(table) do
				if doMath then index = index + 1 end
				if "table" == type(value) then
					for key2, value2 in pairs(value) do
						self:GetPrerequisiteInfo(value2, index)
					end
				else
					self:GetPrerequisiteInfo(value, index)
				end
			end
			return index
		end,

		QuestInfoEnter = function(self, frame)
			self:PopulateTooltipForQuest(frame, self:BreadcrumbQuestId())
			GameTooltip:Show()
		end,

		QuestInfoSection = function(self, heading, table, lastUsedIndex)
			if nil == table then return lastUsedIndex end
			self.tt:AddLine(" ")
			self.tt:AddLine(heading)
			local index = lastUsedIndex or 0
			for key, value in pairs(table) do
				index = index + 1
				if "table" == type(value) then
					for key2, value2 in pairs(value) do
						self.tt:AddDoubleLine("("..index..") "..self:PrettyQuestString({ value2, self:ClassificationOfQuest(value2) }), value2)
					end
				else
					self.tt:AddDoubleLine("("..index..") "..self:PrettyQuestString({ value, self:ClassificationOfQuest(value) }), value)
				end
			end
			return index
		end,

		QuestInfoTurninSection = function(self, heading, table)
			if nil == table then return end
			self.tt:AddLine(" ")
			self.tt:AddLine(heading)
			for key, value in pairs(table) do
				if "table" == type(value) and 2 == #value then
					self.tt:AddDoubleLine(Grail:NPCName(value[1]), self:PrettyQuestString({ value[2], self:ClassificationOfQuest(value[2]) }).." "..value[2])
				else
					self.tt:AddLine("Internal Error with OTC: ", value)
				end
			end
		end,

		QuestName = function(self, questId)
			return Grail:QuestName(questId) or "NO NAME"
		end,

		QuestTypeString = function(self, questId)
			local retval = ""
			local bitValue = Grail:CodeType(questId)
			if bitValue > 0 then
				if bit.band(bitValue, Grail.bitMaskQuestRepeatable) > 0 then retval = retval .. self.s['REPEATABLE'] .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestDaily) > 0 then retval = retval .. DAILY .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestWeekly) > 0 then retval = retval .. CALENDAR_REPEAT_WEEKLY .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestMonthly) > 0 then retval = retval .. CALENDAR_REPEAT_MONTHLY .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestYearly) > 0 then retval = retval .. self.s['YEARLY'] .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestEscort) > 0 then retval = retval .. self.s['ESCORT'] .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestDungeon) > 0 then retval = retval .. CALENDAR_TYPE_DUNGEON .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestRaid) > 0 then retval = retval .. CALENDAR_TYPE_RAID .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestPVP) > 0 then retval = retval .. CALENDAR_TYPE_PVP .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestGroup) > 0 then retval = retval .. CHANNEL_CATEGORY_GROUP .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestHeroic) > 0 then retval = retval .. PLAYER_DIFFICULTY2 .. " " end
			end
			return retval
		end,

		ScrollFrame_OnLoad = function(self, frame)
			HybridScrollFrame_OnLoad(frame)
			frame.update = Wholly.ScrollFrame_Update_WithCombatCheck
			HybridScrollFrame_CreateButtons(frame, "com_mithrandir_whollyButtonTemplate")
		end,

		ScrollFrame_Update_WithCombatCheck = function(self)
			if not InCombatLockdown() then
				Wholly:ScrollFrame_Update()
			else
				com_mithrandir_whollyFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
			end
		end,

		ScrollFrame_Update = function(self)
			self = self or Wholly
			self:FilterPanelQuests()
			local questsInMap = self.filteredPanelQuests
			local numEntries = #questsInMap
			local buttons = com_mithrandir_whollyFrameScrollFrame.buttons
			local numButtons = #buttons
			local scrollOffset = HybridScrollFrame_GetOffset(com_mithrandir_whollyFrameScrollFrame)
			local buttonHeight = buttons[1]:GetHeight();
			local button, questIndex, questId, questLevelString, requiredLevelString, colorCode, questLevel, filterCode, repeatableCompletedString
			local shouldShowTag

			table.sort(questsInMap, Wholly.SortingFunction)
			for i = 1, numButtons do
				button = buttons[i]
				questIndex = i + scrollOffset
				if questIndex <= numEntries then
					questId = questsInMap[questIndex][1]
					filterCode = questsInMap[questIndex][2]
					button.normalText:SetText(self:PrettyQuestString(questsInMap[questIndex]))
					shouldShowTag = false
					if 'I' == filterCode and WhollyDatabase.showsInLogQuestStatus and WhollyDatabase.showsQuestsInLog then
						local questStatus = Grail:StatusCode(questId)
						local statusText = nil
						shouldShowTag = true
						if bit.band(questStatus, Grail.bitMaskInLogComplete) > 0 then statusText = COMPLETE
						elseif bit.band(questStatus, Grail.bitMaskInLogFailed) > 0 then statusText = FAILED
						else shouldShowTag = false
						end
						if nil ~= statusText then
							button.tag:SetText("(" .. statusText .. ")")
						end
					end
					if shouldShowTag then button.tag:Show() else button.tag:Hide() end
					button.questId = questId
					button.statusCode = filterCode
					button:Show()
				else
					button:Hide()
				end
			end
			HybridScrollFrame_Update(com_mithrandir_whollyFrameScrollFrame, numEntries * buttonHeight, numButtons * buttonHeight)
		end,

		SearchEntered = function(self)
			local searchText = com_mithrandir_whollySearchEditBox:GetText()
			com_mithrandir_whollySearchEditBox:SetText("")
			com_mithrandir_whollySearchFrame:Hide()

			-- Remove leading and trailing whitespace
			searchText = searchText:gsub("^%s*(.-)%s*$", "%1")

			if searchText and "" ~= searchText then
				if nil == WhollyDatabase.searches then WhollyDatabase.searches = {} end
				tinsert(WhollyDatabase.searches, searchText)
				if #(WhollyDatabase.searches) > self.maximumSearchHistory then
					tremove(WhollyDatabase.searches, 1)
				end
				self:SearchForQuestNamesMatching(searchText)
				self.currentPanelMapID = 0
				self:ForcePanelMapArea()
			end
		end,

		SearchForQuestNamesMatching = function(self, searchTerm)
			-- the searchTerm is broken up by spaces which are considered AND conditions
			local terms = { strsplit(" ", searchTerm) }
			local results = {}
			local started = false
			for i = 1, #terms do
				if terms[i] ~= "" then
					if not started then
						for q, v in pairs(Grail.quests) do
							if strfind(v[0], terms[i]) then tinsert(results, q) end
						end
						started = true
					else
						local newResults = {}
						for _, q in pairs(results) do
							if strfind(Grail.quests[q][0], terms[i]) then tinsert(newResults, q) end
						end
						results = newResults
					end
				end
			end
			-- clear the mapArea 0 because that is what we use for computed results
			Grail.indexedQuests[0] = {}
			Grail.mapAreaMapping[0] = SEARCH .. ': ' .. searchTerm
			for _, q in pairs(results) do
				Grail:AddQuestToMapArea(q, 0)
			end
		end,

		SetCurrentZone = function(self, frame)
			self:UpdateQuestCaches(false, false, true)
			self:ZoneButtonEnter(frame)	-- need to update the tooltip which is showing
		end,

		ShowBreadcrumbInfo = function(self)
			local questId = self:BreadcrumbQuestId()
			local breadcrumbs = Grail:AvailableBreadcrumbs(questId)
			com_mithrandir_whollyBreadcrumbFrame:Hide()
			com_mithrandir_whollyQuestInfoFrameText:SetText(questId)
			if nil ~= breadcrumbs then
				if 1 == #breadcrumbs then com_mithrandir_whollyBreadcrumbFrameMessage:SetText(self.s['SINGLE_BREADCRUMB_FORMAT'])
				else com_mithrandir_whollyBreadcrumbFrameMessage:SetText(format(self.s['MULTIPLE_BREADCRUMB_FORMAT'], #breadcrumbs))
				end
				com_mithrandir_whollyBreadcrumbFrame:Show()
			end
		end,

		ShowPin = function(self, questTable)
			local codeMapping = { ["G"] = 1, ["W"] = 2, ["D"] = 3, ["R"] = 4, ["P"] = 5, ["L"] = 6, ["U"] = 7 }
			local id = questTable[1]
			local locations = Grail:MultipleUniqueMapLocationQuest(id, 'A', self.currentPinMapID, self.currentPinMapLevel)
			if nil ~= locations then
				for i = 1, #locations do
					local xcoord, ycoord, npcName, npcId = locations[i][1], locations[i][2], locations[i][3], tonumber(locations[i][4])
					if nil ~= xcoord then
						-- record the NPC as needing a tooltip note for the specific quest (it can be a redirect because an actual "NPC" may be the item that starts the quest)
						local shouldProcess, kindsOfNPC = Grail:IsTooltipNPC(npcId)
						if shouldProcess then
							for i = 1, #(kindsOfNPC), 1 do
								local npcIdToUse = npcId
								local shouldAdd = true
								if kindsOfNPC[i][1] == Grail.NPC_TYPE_DROP then
									shouldAdd = self:DroppedItemMatchesQuest(kindsOfNPC[i][2], id)
								end
								if kindsOfNPC[i][1] == Grail.NPC_TYPE_BY then npcIdToUse = tonumber(kindsOfNPC[i][2]) end
								if nil == self.npcs[npcIdToUse] then self.npcs[npcIdToUse] = {} end
								if shouldAdd and not tContains(self.npcs[npcIdToUse], id) then tinsert(self.npcs[npcIdToUse], id) end
							end
						end

						local pin = self:GetPin(npcId)
						local code = questTable[2]
						if 'D' == code and Grail:IsRepeatable(id) then code = 'R' end
						local codeValue = codeMapping[code]
						local pinValue = codeMapping[pin.texType]
						if codeValue < pinValue then
							pin:SetType(code)
						end						
						pin:ClearAllPoints()
						pin.questId = id
						pin:SetFrameLevel(WorldMapPOIFrame:GetFrameLevel() - 1)
						pin:SetPoint("CENTER", WorldMapDetailFrame, "TOPLEFT", xcoord/100*WorldMapDetailFrame:GetWidth(), -ycoord/100*WorldMapDetailFrame:GetHeight())
						pin:Show()

						self.regPins[id..":"..npcId] = pin
					end
				end
			end
		end,

		ShowTooltip = function(self, pin)
			local listedQuests = {}
			self.tooltip:SetOwner(pin, "ANCHOR_RIGHT")
			self.tooltip:ClearLines()
	
			-- find all quests in range of hover
			local mx, my = self:GetMousePosition()
			local npcList = {}
			local npcNames = {}

			local mapId = self.currentPinMapID
			local mapLevel = self.currentPinMapLevel
			local questsInMap = self.filteredPinQuests
			local questId
			for i = 1, #questsInMap do
				questId = questsInMap[i][1]
				local locations, mustKill = Grail:MultipleUniqueMapLocationQuest(questId, 'A', mapId, mapLevel)
				if nil ~= locations then
					for j = 1, #locations do
						local xcoord, ycoord, npcname, npc, areaId, dungeonLevel, realAreaId = locations[j][1], locations[j][2], locations[j][3], locations[j][4], locations[j][5], locations[j][6], locations[j][7]
						if nil ~= xcoord then
							local dist = self:Distance(mx, my, xcoord/100, ycoord/100)
							if (dist <= 0.02) then
								if (not npcList[npc]) then
									npcList[npc] = {}
									npcname = self:PrettyNPCString(npcname, mustKill, realAreaId)
									npcNames[npc] = npcname
								end
								table.insert(npcList[npc], questsInMap[i])
							end
						end
					end
				end
			end

			local first = true;
			for npc,questList in pairs(npcList) do
				if (not first) then
					self.tooltip:AddLine(" ")
				else
					first = false
				end
				for _,qt in ipairs(questList) do
					local leftStr = self:PrettyQuestString(qt)
					local q = qt[1]
					local escortString = Grail:IsEscort(q) and self.s['ESCORT'] or ""
					local dungeonString = Grail:IsDungeon(q) and CALENDAR_TYPE_DUNGEON or ""
					local raidString = Grail:IsRaid(q) and CALENDAR_TYPE_RAID or ""
					local pvpString = Grail:IsPVP(q) and CALENDAR_TYPE_PVP or ""
					local groupString = Grail:IsGroup(q) and CHANNEL_CATEGORY_GROUP or ""
					local heroicString = Grail:IsHeroic(q) and PLAYER_DIFFICULTY2 or ""
					local rightStr = format("%s%s%s%s%s%s", escortString, dungeonString, raidString, pvpString, groupString, heroicString)
					if strlen(rightStr) > 0 then rightStr = format("|c%s%s|r", self.color[qt[2]], rightStr) end

					-- check if already printed - this is for spam quests like the human starting area that haven't been labeled correctly
					if not questName or not listedQuests[questName] then
						self.tooltip:AddDoubleLine(leftStr, rightStr)
						self.tooltip:SetLastFont(self.tooltip.large)
						self.tooltip:SetLastFont(self.tooltip.small, true)
						if questName then listedQuests[questName] = true end
					end
				end
				self.tooltip:AddLine(npcNames[npc], 1, 1, 1, 1)
				self.tooltip:SetLastFont(self.tooltip.small)
			end
	
			self.tooltip:Show();
		end,

		SlashCommand = function(self, frame, msg)
			self:ToggleUI()
		end,

		Sort = function(self)
			-- This is supposed to cycle through the supported sorting techniques and make the contents of the panel
			-- show the quests based on those techniques.
			-- 1 Quest alphabetical
			-- 2 Quest level (and then alphabetical)
			-- 3 Quest type (and then alphabetical)
			WhollyDatabase.currentSortingMode = WhollyDatabase.currentSortingMode + 1
			if (WhollyDatabase.currentSortingMode > 3) then WhollyDatabase.currentSortingMode = 1 end
			self:ScrollFrame_Update_WithCombatCheck()
		end,

		SortingFunction = function(a, b)
			local retval = false
			if 1 == WhollyDatabase.currentSortingMode then
				retval = Wholly:QuestName(a[1]) < Wholly:QuestName(b[1])
			elseif 2 == WhollyDatabase.currentSortingMode then
				local aLevel, bLevel = Grail:QuestLevel(a[1]) or 1, Grail:QuestLevel(b[1]) or 1
				if aLevel == bLevel then
					retval = Wholly:QuestName(a[1]) < Wholly:QuestName(b[1])
				else
					retval = aLevel < bLevel
				end
			elseif 3 == WhollyDatabase.currentSortingMode then
				local aCode, bCode = a[2], b[2]
				if aCode == bCode then
					retval = Wholly:QuestName(a[1]) < Wholly:QuestName(b[1])
				else
					retval = aCode < bCode
				end
			end
			return retval
		end,

		ToggleUI = function(self)
			if com_mithrandir_whollyFrame:IsShown() then
				com_mithrandir_whollyFrame:Hide()
			else
				com_mithrandir_whollyFrame:Show()
			end
		end,

		---
		--	Sets up the event monitoring to handle those associated with displaying breadcrumb information.
		UpdateBreadcrumb = function(self)
			if WhollyDatabase.displaysBreadcrumbs then
				self.notificationFrame:RegisterEvent("QUEST_DETAIL")
				if QuestFrame:IsVisible() then
					self:ShowBreadcrumbInfo()
				end
			else
				self.notificationFrame:UnregisterEvent("QUEST_DETAIL")
				com_mithrandir_whollyBreadcrumbFrame:Hide()
			end
		end,

		UpdateCoordinateSystem = function(self)
			if WhollyDatabase.enablesPlayerCoordinates then
				self.notificationFrame:SetScript("OnUpdate", function(frame, ...) self:OnUpdate(frame, ...) end)
			else
				self.notificationFrame:SetScript("OnUpdate", nil)
				if nil ~= self.coordinates then
					self.coordinates.text = ""
				end
				self.previousX = 0
			end
		end,

		UpdatePins = function(self)
			self:HideAllPins()
			if WhollyDatabase.displaysMapPins then
				self:FilterPinQuests()
				local questsInMap = self.filteredPinQuests
				for i = 1, #questsInMap do
					self:ShowPin(questsInMap[i])
				end
			else
				if 0 == #(self.filteredPinQuests) then self:FilterPinQuests() end
				self.mapCountLine = self:PrettyQuestCountString(self.cachedPinQuests, #(self.filteredPinQuests), true)
			end
		end,

		UpdateQuestCaches = function(self, forceUpdate, setPinMap, setPanelMap)
			if not Grail:IsPrimed() then return end
			local mapArea = GetCurrentMapAreaID()
			local dungeonLevel = GetCurrentMapDungeonLevel() or 0
			if mapArea ~= self.currentPinMapID or dungeonLevel ~= self.currentPinMapLevel or forceUpdate then
				if setPinMap then self.currentPinMapID = mapArea self.currentPinMapLevel = dungeonLevel end
				self.cachedPinQuests = self:ClassifyQuestsInMap(self.currentPinMapID) or {}
				self:UpdatePins()
			end
			if mapArea ~= self.currentPanelMapID or forceUpdate then
				if setPanelMap then self.currentPanelMapID = mapArea end
				self:ForcePanelMapArea()
			end
		end,

		ZoneButtonEnter = function(self, frame)
			GameTooltip:ClearLines()
			GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
			GameTooltip:AddLine(Wholly.panelCountLine)
			GameTooltip:Show()
			GameTooltip:ClearAllPoints()
		end,

		}

	local locale = GetLocale()
	local C = Wholly.configuration
	if "deDE" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Zeige |cFF0099CCwiederholbaren Quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Zeige |cFF00FF00abgeschlossene Quests|r"
		C[6][0]="Zeige |cFF996600unerreichbar Quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Anzeige Pins für Questgeber"
		C[13][0]="Display button on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Display breadcrumb quest information on Quest Frame"
		C[16][0]="Display holiday quests always"
		C[17][0]="Open world map updates when zones change"
		C[18][0]="Show status of quests in log"
		C[19][0]="Show only Loremaster quests"
		C[20][0]="Enable player coordinates"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "Weltereignisse",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['PREQ_EVER'] = "Prerequisite Quests (Ever Completed):",
			['PREQ_LOG'] = "Prerequisite Quests (Must Be in Log):",
			['PREQ_LOG_TURN'] = "Prerequisite Quests (Must Be in Log or Turned in):",
			['PREQ_COMPLETE'] = "Prerequisite Quests (Must Be Completed in Log):",
			['PREQ_COMPLETE_TURN'] = "Prerequisite Quests (Must Be Completed in Log or Turned in):",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			}
	elseif "esES" == locale or "esMX" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show |cFF0099CCrepeatable quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display button on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Display breadcrumb quest information on Quest Frame"
		C[16][0]="Display holiday quests always"
		C[17][0]="Open world map updates when zones change"
		C[18][0]="Show status of quests in log"
		C[19][0]="Show only Loremaster quests"
		C[20][0]="Enable player coordinates"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Matar para iniciar [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Matar]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "Eventos del mundo",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['PREQ_EVER'] = "Prerequisite Quests (Ever Completed):",
			['PREQ_LOG'] = "Prerequisite Quests (Must Be in Log):",
			['PREQ_LOG_TURN'] = "Prerequisite Quests (Must Be in Log or Turned in):",
			['PREQ_COMPLETE'] = "Prerequisite Quests (Must Be Completed in Log):",
			['PREQ_COMPLETE_TURN'] = "Prerequisite Quests (Must Be Completed in Log or Turned in):",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			}
	elseif "frFR" == locale then
		C[1][0]="Ajouter au début du nom de la quête, le niveau de la quête"
		C[2][0]="Append required level"
		C[3][0]="Montrer |cFF0099CCquêtes répétables|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Montrer les |cFF996600quêtes inaccessibles|r"
		C[7][0]="Montrer les |cFF666666quêtes de bas niveau|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Montrer les quêtes de haut niveau"
		C[10][0]="Montrer les |cFFFF0000quêtes qui ont besoin de conditions préalables|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Afficher repères de la carte pour les donneurs de quêtes"
		C[13][0]="Display button on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Display breadcrumb quest information on Quest Frame"
		C[16][0]="Display holiday quests always"
		C[17][0]="Open world map updates when zones change"
		C[18][0]="Show status of quests in log"
		C[19][0]="Show only Loremaster quests"
		C[20][0]="Enable player coordinates"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Tuer pour commencer [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Tuer]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "Évènements mondiaux",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['PREQ_EVER'] = "Prerequisite Quests (Ever Completed):",
			['PREQ_LOG'] = "Prerequisite Quests (Must Be in Log):",
			['PREQ_LOG_TURN'] = "Prerequisite Quests (Must Be in Log or Turned in):",
			['PREQ_COMPLETE'] = "Prerequisite Quests (Must Be Completed in Log):",
			['PREQ_COMPLETE_TURN'] = "Prerequisite Quests (Must Be Completed in Log or Turned in):",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			}
	elseif "koKR" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show |cFF0099CCrepeatable quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display button on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Display breadcrumb quest information on Quest Frame"
		C[16][0]="Display holiday quests always"
		C[17][0]="Open world map updates when zones change"
		C[18][0]="Show status of quests in log"
		C[19][0]="Show only Loremaster quests"
		C[20][0]="Enable player coordinates"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "이벤트",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['PREQ_EVER'] = "Prerequisite Quests (Ever Completed):",
			['PREQ_LOG'] = "Prerequisite Quests (Must Be in Log):",
			['PREQ_LOG_TURN'] = "Prerequisite Quests (Must Be in Log or Turned in):",
			['PREQ_COMPLETE'] = "Prerequisite Quests (Must Be Completed in Log):",
			['PREQ_COMPLETE_TURN'] = "Prerequisite Quests (Must Be Completed in Log or Turned in):",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			}
	elseif "ptBR" == locale then
		C[1][0]="Prefixar nível das missões"
		C[2][0]="Acrescentar nível requerido"
		C[3][0]="Exibir |cFF0099CCmissões repetíveis|r"
		C[4][0]="Exibir |cFFFF00FFmissões no registro|r"
		C[5][0]="Exibir |cFF00FF00missões completadas|r"
		C[6][0]="Exibir |cFF996600missões impossíveis de se obter|r"
		C[7][0]="Exibir |cFF666666missões de nível baixo|r"
		C[8][0]="Mostrar quaisquer missões repetíveis já completadas"
		C[9][0]="Exibir missões de nível alto"
		C[10][0]="Exibir |cFFFF0000missões que precisam de pre-requisitos|r"
		C[11][0]="Painel de registro das missões atualiza quando mudar de zona"
		C[12][0]="Marcar recrutadores no mapa"
		C[13][0]="Exibir botão no mapa mundi"
		C[14][0]="Exibir missões de masmorras no mapa exterior"
		C[15][0]="Mostrar informação sobre missões guia no quadro de missões"
		C[16][0]="Exibir missões de eventos sempre"
		C[17][0]="Abrir atualizações do mapa mundi quando trocar de zona"
		C[18][0]="Exibir status das missões no registro"
		C[19][0]="Exibir somente missões do Mestre Historiador"
		C[20][0]="Enable player coordinates"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Matar para começar [%s]",
			['DROP_TO_START_FORMAT'] = "Saqueie %s para começar [%s]",
			['REQUIRES_FORMAT'] = "Wholly requer a verão %s do Grail ou maior",
			['MUST_KILL_PIN_FORMAT'] = "%s [Matar]",
			['ESCORT'] = "Escolta",
			['BREADCRUMB'] = "Missão guia:",
			['PREREQUISITES'] = "Missões requeridas:",
			['OTHER'] = "Outro",
			['SINGLE_BREADCRUMB_FORMAT'] = "Missão guia disponível",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Missão guia disponível",
			['WORLD_EVENTS'] = "Eventos Mundiais",
			['REPUTATION_REQUIRED'] = "Requer reputação:",
			['REPEATABLE'] = "Repetitível",
			['YEARLY'] = "Anual",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail não tem essa missão|r",
			['QUEST_ID'] = "ID da missão: ",
			['REQUIRED_LEVEL'] = "Requer nível",
			['MAXIMUM_LEVEL_NONE'] = "Nenhum",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "Nenhuma",
			['LOREMASTER_AREA'] = "Área do Mestre Historiador",
			['FACTION_BOTH'] = "Ambas",
			['CLASS_NONE'] = "Nenhuma",
			['CLASS_ANY'] = "Qualquer",
			['GENDER_NONE'] = "Nenhum",
			['GENDER_BOTH'] = "Ambos",
			['GENDER'] = "Gênero",
			['RACE_NONE'] = "Nenhuma",
			['RACE_ANY'] = "Qualquer",
			['HOLIDAYS_ONLY'] = "Disponível apenas durante Eventos Mundiais:",
			['SP_MESSAGE'] = "Missões especiais nunca entram no registro da Blizzard",
			['INVALIDATE'] = "Invalidado pelas missões:",
			['PREQ_EVER'] = "Missões requeridas (Já Concluídas):",
			['PREQ_LOG'] = "Missões requeridas (Devem estar no registro):",
			['PREQ_LOG_TURN'] = "Missões requeridas (devem estar no registro ou entregues):",
			['PREQ_COMPLETE'] = "Missões requeridas (Devem estar completadas no registro):",
			['PREQ_COMPLETE_TURN'] = "Missões requeridas (precisam estar completas no registro ou entregues):",
			['OAC'] = "Após aceitar complete as missões:",
			['OCC'] = "Após completar os requisitos, complete as missões:",
			['OTC'] = "Quando entregue completa a missão:",
			['ENTER_ZONE'] = "Aceita quando entra na área do mapa",
			['WHEN_KILL'] = "Aceita quando matar:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			}
	elseif "ruRU" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Показать |cFF0099CCповторяющиеся квесты|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display button on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Display breadcrumb quest information on Quest Frame"
		C[16][0]="Display holiday quests always"
		C[17][0]="Open world map updates when zones change"
		C[18][0]="Show status of quests in log"
		C[19][0]="Show only Loremaster quests"
		C[20][0]="Enable player coordinates"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Убить, чтобы начать [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Убить]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "Игровые события",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['PREQ_EVER'] = "Prerequisite Quests (Ever Completed):",
			['PREQ_LOG'] = "Prerequisite Quests (Must Be in Log):",
			['PREQ_LOG_TURN'] = "Prerequisite Quests (Must Be in Log or Turned in):",
			['PREQ_COMPLETE'] = "Prerequisite Quests (Must Be Completed in Log):",
			['PREQ_COMPLETE_TURN'] = "Prerequisite Quests (Must Be Completed in Log or Turned in):",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			}
	elseif "zhCN" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show |cFF0099CCrepeatable quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display button on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Display breadcrumb quest information on Quest Frame"
		C[16][0]="Display holiday quests always"
		C[17][0]="Open world map updates when zones change"
		C[18][0]="Show status of quests in log"
		C[19][0]="Show only Loremaster quests"
		C[20][0]="Enable player coordinates"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['PREQ_EVER'] = "Prerequisite Quests (Ever Completed):",
			['PREQ_LOG'] = "Prerequisite Quests (Must Be in Log):",
			['PREQ_LOG_TURN'] = "Prerequisite Quests (Must Be in Log or Turned in):",
			['PREQ_COMPLETE'] = "Prerequisite Quests (Must Be Completed in Log):",
			['PREQ_COMPLETE_TURN'] = "Prerequisite Quests (Must Be Completed in Log or Turned in):",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			}
	elseif "zhTW" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show |cFF0099CCrepeatable quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display button on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Display breadcrumb quest information on Quest Frame"
		C[16][0]="Display holiday quests always"
		C[17][0]="Open world map updates when zones change"
		C[18][0]="Show status of quests in log"
		C[19][0]="Show only Loremaster quests"
		C[20][0]="Enable player coordinates"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "世界事件",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['PREQ_EVER'] = "Prerequisite Quests (Ever Completed):",
			['PREQ_LOG'] = "Prerequisite Quests (Must Be in Log):",
			['PREQ_LOG_TURN'] = "Prerequisite Quests (Must Be in Log or Turned in):",
			['PREQ_COMPLETE'] = "Prerequisite Quests (Must Be Completed in Log):",
			['PREQ_COMPLETE_TURN'] = "Prerequisite Quests (Must Be Completed in Log or Turned in):",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			}
	end

end
